#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
__version__ = "1.0.13"

import os
import re
import sys
import logging
from typing import Optional, List, Dict, Any

from mlplatform_client.clients import BaseClient
from mlplatform_client.core.settings import APP_SETTINGS
from mlplatform_client.core.manifest import Manifest
from mlplatform_client.core.decorators import access_token, validate_token
from mlplatform_client.serializers.auth import TokenRegisterIn, MLPTokenType, DeleteTokenPayload, TokenInfo
from mlplatform_client.clients.auth_client import AuthClient
from mlplatform_client.clients.spark_client import SparkClient
from mlplatform_client.clients.gateway_client import GatewayClient
from mlplatform_client.clients.deploy_client import DeployServerClient
from mlplatform_client.clients.model_client import DeployModelClient
from mlplatform_client.core.utils import create_uq_str_with_prefix
from mlplatform_client.serializers.deployment import DeployOut, DeploymentsList, UserAuth
from mlplatform_client.serializers.gateway import (
    VmDetailInfo,
    MLPlatformIN,
    MLPlatformVolumeIn,
    MLPlatformNetworkIn,
    InstanceType,
    BucketIn,
    NetworkOut,
    FrontInstanceStatus,
)
from mlplatform_client.serializers.nova import NovaFlavor
from mlplatform_client.serializers.spark_proxy import (
    K8sSparkJobInfo,
    K8sSparkJobInfoDetail,
    K8sSparkJobLogs,
    K8sSparkEvent,
    K8sSparkJobIn,
    K8sClusterInfo,
    K8sSparkEventsList,
    K8sSparkJobsInfoList,
    K8sSparkApplication,
    K8sSecret,
    K8sConfigmap,
    K8sManifestIn,
    K8sObjectNamesList,
    K8sInactiveTimeoutIn,
)
from mlplatform_client.core.exceptions import ConflictException


job_name_validator = re.compile(r'^[a-z0-9][a-z0-9-.]*[a-z0-9]$')
sys.tracebacklimit = 0


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [ %(levelname)-7.7s ]  %(message)s",
    handlers=[logging.StreamHandler()],
)
log = logging.getLogger()


class MLPlatform:
    def __init__(self, refresh_token: Optional[str] = None):
        self._auth_client = AuthClient(config=APP_SETTINGS.auth, refresh_token=refresh_token)
        self._spark_client = SparkClient(config=APP_SETTINGS.spark)
        self._gateway_client = GatewayClient(config=APP_SETTINGS.gateway)

    def stop(self):
        """Stop connection with clients"""
        self._auth_client.stop()
        self._spark_client.stop()
        self._gateway_client.stop()

    @staticmethod
    def _read_file(file_name: str) -> str:
        with open(file_name, "r") as f:
            return f.read()

    @access_token()
    def _get_active_instance_by_name(self, instance_name: str, **kwargs) -> VmDetailInfo:
        instance = self._gateway_client.get_instance_by_name(instance_name=instance_name, **kwargs)
        if instance.status != FrontInstanceStatus.RUNNING:
            raise ConflictException(f"Server '{instance_name}' does not RUNNING")
        if not instance.domain_name:
            raise ConflictException(f"Server '{instance_name}' has no private IP")
        return instance

    @staticmethod
    def get_manifest_from_yaml_file(yaml_file_path: str) -> Manifest:
        """Convert yaml file to Manifest

        Args:
            yaml_file_path (str): path to yaml file

        Returns:
            Manifest
        """

        return Manifest.load_yaml(yaml_file_path)

    @access_token(throw_access_token=False)
    def who_am_i(self) -> TokenInfo:
        """Get info about token

        Returns:
            TokenInfo: Token info
        """
        return self._auth_client.who_am_i()

    @access_token(throw_access_token=False)
    def get_project_tokens_info(self) -> List[TokenInfo]:
        """Get info about project tokens

        Returns:
            List[TokenInfo]: Project tokens info
        """
        return self._auth_client.get_project_tokens_info()

    @access_token(throw_access_token=False)
    def create_register_token(
            self, client_name: str,
            access_ttl: Optional[str],
            refresh_ttl: str = None,
            token_type: MLPTokenType = MLPTokenType.USER,
    ) -> str:
        """Creates a register token with admin or user rights, token allows to create refresh token. The register token
        for other admins and users is allowed to be created only by users with admin rights.

        Example:
            from mlplatform_client.serializers.auth import MLPTokenType

            create_register_token(
                client_name="client_name", access_ttl="2h", refresh_ttl="720h", token_type=MLPTokenType.ADMIN
            )

        Args:
            client_name (str): name in any format
            access_ttl (Optional[str]): access token lifetime. There can be seconds, minutes, hours and their
                combinations. Valid time units are "s", "m", "h". For example "300s", "2h", "2h45m"
            refresh_ttl (str): refresh token lifetime. Equivalent to access_ttl
            token_type (MLPTokenType): can be MLPTokenType.USER or MLPTokenType.ADMIN, default MLPTokenType.USER
            **kwargs:

        Returns:
            str: string which contains register token
        """
        payload = TokenRegisterIn(
            client_name=client_name,
            token_type=token_type,
            access_ttl=access_ttl,
            refresh_ttl=refresh_ttl,
        )
        return self._auth_client.create_register_token(payload=payload)

    @validate_token()
    def create_refresh_token(self, register_token: str) -> str:
        """Creates a refresh token, needed to perform all actions on the MLPlatform

        Args:
            register_token (str): register token

        Returns:
            str: string which contains refresh token
        """
        return self._auth_client.create_refresh_token(register_token)

    @access_token(throw_access_token=False)
    def delete_token(self, token_id: Optional[str] = None, token_name: Optional[str] = None) -> None:
        """Delete token by parameters.

        Args:
            token_id (Optional[str]): UUID of deletion token
            token_name (Optional[str]): Name of deletion token
        """
        req = DeleteTokenPayload(token_id=token_id, token_name=token_name)
        return self._auth_client.delete_token(req)

    @access_token()
    def get_clusters(self, **kwargs) -> List[K8sClusterInfo]:
        """Get all available k8s clusters

        Args:
            **kwargs:

        Returns:
            List[K8sClusterInfo]: list with k8s clusters info
        """
        return self._spark_client.get_clusters(**kwargs)

    @access_token()
    def get_cluster_info(self, cluster_id: str, **kwargs) -> K8sClusterInfo:
        """Get info about k8s cluster

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            **kwargs:

        Returns:
            K8sClusterInfo: info about k8s cluster
        """
        return self._spark_client.get_cluster_info(cluster_id=cluster_id, **kwargs)

    @access_token()
    def set_cluster_suspend_time(self, cluster_id: str, minutes: int, **kwargs) -> K8sClusterInfo:
        """Sets cluster inactivity timeout after which cluster will be automatically suspended until woken up

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            minutes (int): Inactivity time in minutes after which cluster will be suspended automatically
            **kwargs:

        Returns:
            K8sClusterInfo: info about k8s cluster
        """
        payload = K8sInactiveTimeoutIn(suspend_timeout_min=minutes)
        return self._spark_client.set_cluster_timeout(cluster_id=cluster_id, payload=payload, **kwargs)

    @access_token()
    def set_cluster_delete_time(self, cluster_id: str, minutes: int, **kwargs) -> K8sClusterInfo:
        """Sets cluster inactivity timeout after which cluster will be automatically DELETED permanently

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            minutes (int): Inactivity time in minutes after which cluster will be DELETED permanently
            **kwargs:

        Returns:
            K8sClusterInfo: info about k8s cluster
        """
        log.warning(f"!!! Cluster will be permanently DELETED after {minutes} minutes of inactivity !!!")
        payload = K8sInactiveTimeoutIn(delete_timeout_min=minutes)
        return self._spark_client.set_cluster_timeout(cluster_id=cluster_id, payload=payload, **kwargs)

    @access_token()
    def remove_cluster_suspend_time(self, cluster_id: str, **kwargs) -> K8sClusterInfo:
        """Removes suspend inactivity timeout for cluster

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            **kwargs:

        Returns:
            K8sClusterInfo: info about k8s cluster
        """
        payload = K8sInactiveTimeoutIn(suspend_timeout_min=0)
        return self._spark_client.set_cluster_timeout(cluster_id=cluster_id, payload=payload, **kwargs)

    @access_token()
    def remove_cluster_delete_time(self, cluster_id: str, **kwargs) -> K8sClusterInfo:
        """Removes delete inactivity timeout for cluster

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            **kwargs:

        Returns:
            K8sClusterInfo: info about k8s cluster
        """
        payload = K8sInactiveTimeoutIn(delete_timeout_min=0)
        return self._spark_client.set_cluster_timeout(cluster_id=cluster_id, payload=payload, **kwargs)

    @access_token()
    def spark_submit_job(
            self, cluster_id: str, manifest: Manifest, pycode_file_path: Optional[str] = None, **kwargs
    ) -> K8sSparkJobInfo:
        """Submit spark job to k8s cluster

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            manifest (Manifest): manifest class containing k8s spark operator manifest
            pycode_file_path (str): path to python code file, which pass to k8s cluster to use in spark job
            **kwargs:

        Returns:
            K8sSparkJobInfo: info about spark job
        """
        if not job_name_validator.fullmatch(manifest.job_name):
            raise ValueError(
                f"Job_name {manifest.job_name} does not comply with RFC 1123 subdomain. Allowed symbols a-z, 0-9, '-', '.', and must start and end with an alphanumeric character"
            )
        pycode = None
        pycode_file_name = None
        if pycode_file_path is not None:
            pycode = self._read_file(pycode_file_path)
            _, pycode_file_name = os.path.split(pycode_file_path)
        payload = K8sSparkJobIn(
            spark_manifest=manifest.to_yaml(),
            python_file_name=pycode_file_name,
            python_code=pycode,
        )

        return self._spark_client.submit_job(cluster_id=cluster_id, payload=payload, **kwargs)
    
    @access_token()
    def spark_job_info(self, cluster_id: str, job_name: str, **kwargs) -> K8sSparkJobInfoDetail:
        """Get detail info of spark job

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)
            **kwargs:

        Returns:
            K8sSparkJobInfoDetail: detail info about spark job
        """
        return self._spark_client.job_info(job_name=job_name, cluster_id=cluster_id, **kwargs)

    @access_token()
    def spark_jobs_list(self, cluster_id: str, **kwargs) -> K8sSparkJobsInfoList:

        """Get list of spark jobs info

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            **kwargs:

        Returns:
            K8sSparkJobsInfoList: list with info about spark jobs
        """
        return self._spark_client.jobs_list(cluster_id=cluster_id, **kwargs)

    @access_token()
    def spark_job_logs(self, cluster_id: str, job_name: str, **kwargs) -> K8sSparkJobLogs:
        """Get logs of spark job

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)
            **kwargs:

        Returns:
            K8sSparkJobLogs: logs of spark job
        """
        return self._spark_client.job_logs(job_name=job_name, cluster_id=cluster_id, **kwargs)

    @access_token()
    def describe_spark_job(self, cluster_id: str, job_name: str, **kwargs) -> K8sSparkApplication:
        """Describes job by its name

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)
            **kwargs:

        Returns:
            K8sSparkApplication: description of spark job (application)
        """
        return self._spark_client.describe_job(job_name=job_name, cluster_id=cluster_id, **kwargs)

    @access_token()
    def spark_events(self, cluster_id: str, **kwargs) -> K8sSparkEventsList:

        """Get events of spark cluster

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            **kwargs:

        Returns:
            K8sSparkEventsList: list of spark cluster events
        """
        return self._spark_client.events(cluster_id=cluster_id, **kwargs)

    @access_token()
    def spark_delete_job(self, cluster_id: str, job_name: str, **kwargs) -> None:
        """Delete spark job

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)
            **kwargs:
        """
        return self._spark_client.delete_job(cluster_id=cluster_id, job_name=job_name, **kwargs)

    @access_token()
    def get_default_manifest(self, cluster_id: str, job_name: Optional[str] = None, **kwargs) -> Manifest:
        """Get SparkApplication default manifest

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name: job name from the yaml manifest (metadata.name)
            **kwargs:

        Returns:
            Manifest: get example of SparkApplication default manifest
        """
        if job_name and not job_name_validator.fullmatch(job_name):
            raise ValueError(
                f"Job_name {job_name} does not comply with RFC 1123 subdomain. Allowed symbols a-z, 0-9, '-', '.', and must start and end with an alphanumeric character"
            )
        default_manifest = self._spark_client.get_default_manifest(cluster_id=cluster_id, job_name=job_name, **kwargs)
        return Manifest.new_from_yaml(default_manifest)

    @access_token()
    def wake_up_cluster(self, cluster_id: str, **kwargs) -> None:
        """Activate k8s cluster

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            **kwargs:

        Returns:
            None
        """
        return self._gateway_client.wake_up_cluster(cluster_id=cluster_id, **kwargs)

    @access_token()
    def create_secret_from_yaml(
            self, cluster_id: str, secret_yaml_path: str, namespace: str = "default", **kwargs
    ) -> K8sSecret:
        """Create k8s secret from yaml manifest

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            secret_yaml_path (str): path to secret manifest yaml file, which pass to k8s cluster to apply
            namespace (str): path to python code file, which pass to k8s cluster to use in spark job
            **kwargs:

        Returns:
            K8sSecret: info about created k8s secret
        """
        secret_manifest = self._read_file(secret_yaml_path)
        payload = K8sManifestIn(
            manifest=secret_manifest,
        )

        return self._spark_client.create_secret(cluster_id=cluster_id, payload=payload, namespace=namespace, **kwargs)

    @access_token()
    def get_secret_detail(self, cluster_id: str, secret_name: str, namespace: str = "default", **kwargs) -> K8sSecret:
        """Get detail info of secret

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            secret_name (str): name of a secret on k8s cluster
            namespace (Optional[str]): namespace of a secret
            **kwargs:

        Returns:
            K8sSecret: detail info about k8s secret
        """
        return self._spark_client.get_secret(secret_name=secret_name, cluster_id=cluster_id, namespace=namespace, **kwargs)

    @access_token()
    def list_secrets(self, cluster_id: str, namespace: str = "default", **kwargs) -> K8sObjectNamesList[str]:
        """Get names of all secrets for a namespace

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            namespace (Optional[str]): namespace of a configMap
            **kwargs:

        Returns:
            K8sObjectNamesList[str]: k8s secret names
        """
        return self._spark_client.list_secrets(cluster_id=cluster_id, namespace=namespace, **kwargs)

    @access_token()
    def delete_secret(self, cluster_id: str, secret_name: str, namespace: str = "default", **kwargs):
        """Delete k8s secret

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            secret_name (str): name of a secret on k8s cluster
            namespace (Optional[str]): namespace of a secret
            **kwargs:

        """
        return self._spark_client.delete_secret(secret_name=secret_name, cluster_id=cluster_id, namespace=namespace, **kwargs)

    @access_token()
    def create_configmap_from_yaml(
            self, cluster_id: str, configmap_yaml_path: str, namespace: str = "default", **kwargs
    ) -> K8sConfigmap:
        """Create k8s configMap from yaml manifest

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            configmap_yaml_path (str): path to configMap manifest yaml file, which pass to k8s cluster to apply
            namespace (str): path to python code file, which pass to k8s cluster to use in spark job
            **kwargs:

        Returns:
            K8sSecret: info about created k8s configMap
        """
        configmap_manifest = self._read_file(configmap_yaml_path)
        payload = K8sManifestIn(
            manifest=configmap_manifest,
        )

        return self._spark_client.create_configmap(cluster_id=cluster_id, payload=payload, namespace=namespace, **kwargs)

    @access_token()
    def get_configmap_detail(self, cluster_id: str, configmap_name: str, namespace: str = "default", **kwargs) -> K8sConfigmap:
        """Get detail info of configMap

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            configmap_name (str): name of a configMap on k8s cluster
            namespace (Optional[str]): namespace of a configMap
            **kwargs:

        Returns:
            K8sConfigmap: detail info about k8s configMap
        """
        return self._spark_client.get_configmap(configmap_name=configmap_name, cluster_id=cluster_id, namespace=namespace, **kwargs)

    @access_token()
    def list_configmaps(self, cluster_id: str, namespace: str = "default", **kwargs) -> K8sObjectNamesList[str]:
        """Get names of all configMaps for a namespace

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            namespace (Optional[str]): namespace of a configMap
            **kwargs:

        Returns:
            K8sObjectNamesList[str]: k8s configMaps names
        """
        return self._spark_client.list_configmaps(cluster_id=cluster_id, namespace=namespace, **kwargs)

    @access_token()
    def delete_configmap(self, cluster_id: str, configmap_name: str, namespace: str = "default", **kwargs):
        """Delete k8s configMap

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            configmap_name (str): name of a configMap on k8s cluster
            namespace (Optional[str]): namespace of a configMap
            **kwargs:

        """
        return self._spark_client.delete_configmap(configmap_name=configmap_name, cluster_id=cluster_id, namespace=namespace, **kwargs)

    @access_token()
    def get_flavors(self, **kwargs) -> List[NovaFlavor]:
        """[ADMIN] Get all flavors for instances (JupiterHub/MLFlow/Deploy)

        Args:
            **kwargs:

        Returns:
            List[NovaFlavor]: list with flavors info
        """
        return self._gateway_client.get_flavors(**kwargs)

    @access_token()
    def get_all_instances_info(self, **kwargs) -> List[VmDetailInfo]:
        """[ADMIN] Get all instances info (JupiterHub/MLFlow/Deploy)

                Args:
                    **kwargs:

                Returns:
                    List[VmDetailInfo]: list with instances info (JupiterHub/MLFlow/Deploy)
                """
        return self._gateway_client.get_all_instances_info(**kwargs)

    @access_token()
    def get_instance_info(self, instance_id: str, **kwargs) -> VmDetailInfo:
        """[ADMIN] Get instance info (JupiterHub/MLFlow/Deploy)

                Args:
                    instance_id (str): id of instance (JupiterHub/MLFlow/Deploy)
                    **kwargs:

                Returns:
                    VmDetailInfo: instance info (JupiterHub/MLFlow/Deploy)
                """
        return self._gateway_client.get_instance_info(instance_id=instance_id, **kwargs)

    @access_token()
    def create_jupiter_hub(
            self, instance_name: str, jh_admin_name: str, jh_admin_password: str, flavor: str,
            volumes: List[MLPlatformVolumeIn], networks: MLPlatformNetworkIn, domain_name: Optional[str] = None,
            s3fs_bucket: Optional[str] = None, **kwargs
    ) -> VmDetailInfo:
        """[ADMIN] Create JupiterHub

        Example:
            create_jupiter_hub(
                instance_name="instance_name", domain_name="mlp-3c96a8acb40298",
                jh_admin_name="admin", jh_admin_password="Admin123@",flavor="366704e5-2315-4164-bb27-c353dcc2cbbb", volumes=[
                    MLPlatformVolumeIn(size=50, volume_type=VolumeType.ceph_ssd, availability_zone=AvailabilityZone.GZ1)
                ], networks=MLPlatformNetworkIn(network_id="1234")
            )

        Args:
            instance_name (str): name of JupiterHub
            domain_name (Optional[str]): domain name if not specified it will be created automatically.
                The maximum length is 63 characters. Only numbers, lowercase Latin letters, "." and "-" characters
                are allowed. It cannot start and end with "-"
            jh_admin_name (str): jh_admin_name
            jh_admin_password (str): JupiterHub admin password have to more than 8 simbols, contain Latin letters,
                upper case, lower case, numbers and simbols !#$%&()*+,.:;<=>?@[]^_{|}~- For example: Admin123@
            flavor (str): flavor id can get from "get_flavors" method
            volumes (List[MLPlatformVolumeIn]): volumes. For example: [
                MLPlatformVolumeIn(
                    size=50, in Gb
                    volume_type=VolumeType.ceph_ssd, type of SSD can be ceph_ssd or high_iops
                    availability_zone=AvailabilityZone.GZ1 can be GZ1 or MS1
                )
            ]
            networks (MLPlatformNetworkIn): networks. For example: MLPlatformNetworkIn(network_id="12345")
            s3fs_bucket (Optional[str]): you can connect an existing bucket or create a new one by the passed name.
                The name of the bucket must be between 4 and 63 characters long and can only contain lowercase
                Latin letters or numbers. You can only specify a name that no one else has: it must be unique
                for the entire service. Characters are allowed "." or "-" or "_".
                It just can start and end with lowercase Latin letters or numbers.
            **kwargs:

        Returns:
            VmDetailInfo: detail about JupiterHub instance
        """
        if not domain_name:
            domain_name = create_uq_str_with_prefix()

        payload = MLPlatformIN(
            instance_name=instance_name,
            domain_name=domain_name,
            jh_admin_name=jh_admin_name,
            jh_admin_password=jh_admin_password,
            flavor=flavor,
            volumes=volumes,
            networks=networks,
            instance_type=InstanceType.JUPYTERHUB,
            s3fs_bucket=s3fs_bucket,
        )
        return self._gateway_client.create_instance(payload=payload, **kwargs)

    @access_token()
    def delete_instance(self, instance_id: str, **kwargs) -> None:
        """[ADMIN] Delete instance by id (JupiterHub/MLFlow/Deploy)

        Args:
            instance_id (str): id of instance (JupiterHub/MLFlow/Deploy)
            **kwargs:

        Returns:
            None
        """
        return self._gateway_client.delete_instance(instance_id=instance_id, **kwargs)

    @access_token()
    def attach_s3_bucket_to_jh(self, jh_id: str, s3fs_bucket: str, **kwargs):
        """[ADMIN] Attach S3 bucket Jupiter Hubs by id and bucket name

        Args:
            jh_id (str): id of instance Jupiter Hub
            s3fs_bucket (str): you can connect an existing bucket or create a new one by the passed name.
                The name of the bucket must be between 4 and 63 characters long and can only contain lowercase
                Latin letters or numbers. You can only specify a name that no one else has: it must be unique
                for the entire service. Characters are allowed "." or "-" or "_".
                It just can start and end with lowercase Latin letters or numbers.

        Returns:
            None
        """
        payload = BucketIn(
            bucket_name=s3fs_bucket
        )
        return self._gateway_client.attach_s3(jh_id=jh_id, payload=payload, **kwargs)

    @access_token()
    def create_mlflow(
            self, instance_name: str, jh_id: str, flavor: str, volumes: List[MLPlatformVolumeIn],
            networks: MLPlatformNetworkIn, domain_name: Optional[str] = None, is_mlflow_demo_mode=True, **kwargs
    ) -> VmDetailInfo:
        """[ADMIN] Create MLFlow

        Example:
            create_mlflow(
                instance_name="instance_name", jh_id="12345", domain_name="mlp-3c96a8acb40298",
                flavor="366704e5-2315-4164-bb27-c353dcc2cbbb", volumes=[
                    MLPlatformVolumeIn(size=50, volume_type=VolumeType.ceph_ssd, availability_zone=AvailabilityZone.GZ1)
                ], networks=MLPlatformNetworkIn(network_id="1234"), is_mlflow_demo_mode=True
            )

        Args:
            instance_name (str): name of MLFlow
            jh_id (str): Exists Jupiter Hub id for connect with MLFlow
            domain_name (Optional[str]): domain name if not specified it will be created automatically.
                The maximum length is 63 characters. Only numbers, lowercase Latin letters, "." and "-" characters
                are allowed. It cannot start and end with "-"
            flavor (str): flavor id can get from "get_flavors" method
            volumes (List[MLPlatformVolumeIn]): volumes. For example: [
                MLPlatformVolumeIn(
                    size=50, in Gb
                    volume_type=VolumeType.ceph_ssd, type of SSD can be ceph_ssd or high_iops
                    availability_zone=AvailabilityZone.GZ1 can be GZ1 or MS1
                )
            ]
            networks (MLPlatformNetworkIn): networks. For example: MLPlatformNetworkIn(network_id="12345")
            is_mlflow_demo_mode (bool): If True - Demo mode creates MLFlow with local SQLite database, with False
                creates PostgreSQL database
            **kwargs:

        Returns:
            VmDetailInfo: detail about MLFlow instance
        """
        if not domain_name:
            domain_name = create_uq_str_with_prefix()

        payload = MLPlatformIN(
            instance_name=instance_name,
            domain_name=domain_name,
            flavor=flavor,
            volumes=volumes,
            networks=networks,
            instance_type=InstanceType.MLFLOW,
            mlflow_jh_instance_id=jh_id,
            is_mlflow_demo_mode=is_mlflow_demo_mode,
        )
        return self._gateway_client.create_instance(payload=payload, **kwargs)

    @access_token()
    def create_standalone_mlflow(
            self, instance_name: str, mlflow_admin_name: str, mlflow_admin_password: str, flavor: str,
            volumes: List[MLPlatformVolumeIn], networks: MLPlatformNetworkIn, domain_name: Optional[str] = None, **kwargs
    ) -> VmDetailInfo:
        """[ADMIN] Create Mlflow Standalone

        Example:
            create_mlflow_standalone(
                instance_name="instance_name", domain_name="mlp-3c96a8acb40298",
                mlflow_admin_name="admin", mlflow_admin_password="Admin123@",flavor="366704e5-2315-4164-bb27-c353dcc2cbbb", volumes=[
                    MLPlatformVolumeIn(size=50, volume_type=VolumeType.ceph_ssd, availability_zone=AvailabilityZone.GZ1)
                ], networks=MLPlatformNetworkIn(network_id="1234")
            )

        Args:
            instance_name (str): name of Mlflow Standalone
            domain_name (Optional[str]): domain name if not specified it will be created automatically.
                The maximum length is 63 characters. Only numbers, lowercase Latin letters, "." and "-" characters
                are allowed. It cannot start and end with "-"
            mlflow_admin_name (str): mlflow_admin_name
            mlflow_admin_password (str): Mlflow Standalone admin password have to more than 8 simbols, contain Latin letters,
                upper case, lower case, numbers and simbols !#$%&()*+,.:;<=>?@[]^_{|}~- For example: Admin123@
            flavor (str): flavor id can get from "get_flavors" method
            volumes (List[MLPlatformVolumeIn]): volumes. For example: [
                MLPlatformVolumeIn(
                    size=50, in Gb
                    volume_type=VolumeType.ceph_ssd, type of SSD can be ceph_ssd or high_iops
                    availability_zone=AvailabilityZone.GZ1 can be GZ1 or MS1
                )
            ]
            networks (MLPlatformNetworkIn): networks. For example: MLPlatformNetworkIn(network_id="12345")
            **kwargs:

        Returns:
            VmDetailInfo: detail about Mlflow Standalone instance
        """
        if not domain_name:
            domain_name = create_uq_str_with_prefix()

        payload = MLPlatformIN(
            instance_name=instance_name,
            domain_name=domain_name,
            jh_admin_name=mlflow_admin_name,
            jh_admin_password=mlflow_admin_password,
            flavor=flavor,
            volumes=volumes,
            networks=networks,
            instance_type=InstanceType.MLFLOW_STANDALONE,
        )
        return self._gateway_client.create_instance(payload=payload, **kwargs)

    @access_token()
    def create_deploy(
            self, instance_name: str, username: str, password: str,
            mlflow_id: str, flavor: str, volumes: List[MLPlatformVolumeIn],
            networks: MLPlatformNetworkIn, domain_name: Optional[str] = None, is_mlflow_demo_mode=True, **kwargs
    ) -> VmDetailInfo:
        """[ADMIN] Create Deploy server

        Example:
            create_deploy(
                instance_name="instance_name", mlflow_id="12345", domain_name="mlp-3c96a8acb40298",
                flavor="366704e5-2315-4164-bb27-c353dcc2cbbb", volumes=[
                    MLPlatformVolumeIn(size=50, volume_type=VolumeType.ceph_ssd, availability_zone=AvailabilityZone.GZ1)
                ], networks=MLPlatformNetworkIn(network_id="1234"), is_mlflow_demo_mode=True
            )

        Args:
            instance_name (str): name of MLFlow
            username (str): admin name for JupyterHub or MLflow Standalone
            password (str): admin password for JupyterHub or Mlflow Standalone
            mlflow_id (str): Exists MLFlow id for connect with MLFlow
            domain_name (Optional[str]): domain name if not specified it will be created automatically.
                The maximum length is 63 characters. Only numbers, lowercase Latin letters, "." and "-" characters
                are allowed. It cannot start and end with "-"
            flavor (str): flavor id can get from "get_flavors" method
            volumes (List[MLPlatformVolumeIn]): volumes. For example: [
                MLPlatformVolumeIn(
                    size=50, in Gb
                    volume_type=VolumeType.ceph_ssd, type of SSD can be ceph_ssd or high_iops
                    availability_zone=AvailabilityZone.GZ1 can be GZ1 or MS1
                )
            ]
            networks (MLPlatformNetworkIn): networks. For example: MLPlatformNetworkIn(network_id="12345")
            is_mlflow_demo_mode (bool): If True - Demo mode creates MLFlow with local SQLite database, with False
                creates PostgreSQL database
            **kwargs:

        Returns:
            VmDetailInfo: detail about MLFlow instance
        """
        if not domain_name:
            domain_name = create_uq_str_with_prefix()

        payload = MLPlatformIN(
            instance_name=instance_name,
            domain_name=domain_name,
            jh_admin_name=username,
            jh_admin_password=password,
            flavor=flavor,
            volumes=volumes,
            networks=networks,
            instance_type=InstanceType.DEPLOY,
            deploy_mlflow_instance_id=mlflow_id,
            is_mlflow_demo_mode=is_mlflow_demo_mode,
        )
        return self._gateway_client.create_instance(payload=payload, **kwargs)

    @access_token()
    def get_internal_networks(self, **kwargs) -> List[NetworkOut]:
        """[ADMIN] Get all internal networks info

        Args:
            **kwargs:

        Returns:
            List[NetworkOut]: list with internal networks info
        """
        return self._gateway_client.get_internal_networks(**kwargs)

    @access_token()
    def get_external_networks(self, **kwargs) -> List[NetworkOut]:
        """[ADMIN] Get all external networks info

        Args:
            **kwargs:

        Returns:
            List[NetworkOut]: list with external networks info
        """
        return self._gateway_client.get_external_networks(**kwargs)

    @access_token()
    def create_deployment(
            self, deploy_server_name: str, deployment_name: str, model_uri: str, username: str, password: str,
            model_username: str, model_password: str, port_out: int = None, **kwargs
    ) -> DeployOut:
        """Creates model deployment for remote inference

        Args:
            deploy_server_name (str): Name of deploy server
            deployment_name (str): remote model deployment name
            model_uri (str): MLFlow model URI from artifact storage
            username (str): admin name for JupyterHub or MLflow Standalone
            password (str): admin password for JupyterHub or Mlflow Standalone
            port_out (str): Remote model port
            model_username (str): String containing auth information to protect remote model from unauthorized access
                Do not confuse with JupyterHub or MLflow Standalone username as latest used to create deployments
                and model_username used to access remote inferece e.g. 'predict' or 'ping' methods
            model_password (str): Password for remote model inference.
                Do not confuse with JupyterHub or MLflow Standalone passwords as latest used to create deployments
                and model_password used to access remote inferece e.g. 'predict' or 'ping' methods

        Returns:
            DeployOut: Information about model deployments(URI, status, etc.)
        """
        deploy_auth = UserAuth(username=username, password=password)
        model_auth = UserAuth(username=model_username, password=model_password)
        deploy_server = self._get_active_instance_by_name(instance_name=deploy_server_name, **kwargs)
        if not model_uri or not model_uri.startswith("mlflow-artifacts"):
            raise ValueError(
                f"Not full path to model provided {model_uri}. Provide Full Path, e.g. "
                f"mlflow-artifacts:/0/7ffadc140d9e449c89711ed6203243a8/artifacts/model"
            )
        if not deployment_name:
            raise ValueError("param: 'deployment_name' must not be empty")

        deploy_cli = DeployServerClient(deploy_server.domain_name, deploy_auth)
        deploy_list = deploy_cli.list_deployments()
        if DeployServerClient.deployment_exist(deploy_list, deployment_name):
            raise ConflictException(f"Deployment with name: '{deployment_name}' already exist")

        port_out = DeployServerClient.find_next_port(deploy_list, port_out)
        log.warning("Time-consuming operation was started, need to wait 2-3 mins or above")

        deploy = deploy_cli.create_deployment(deployment_name, model_uri, port_out, model_auth)
        return deploy

    @access_token()
    def list_deployments(self, deploy_server_name: str, username: str, password: str, **kwargs) -> List[DeployOut]:
        """Lists all available model deployments

        Args:
            deploy_server_name (str): Name of deploy server
            username (str): admin name for JupyterHub or MLflow Standalone
            password (str): admin password for JupyterHub or Mlflow Standalone

        Returns:
            List[DeployOut]: List of model deployments information(URI, status, etc.)
        """
        deploy_auth = UserAuth(username=username, password=password)
        deploy_server = self._get_active_instance_by_name(instance_name=deploy_server_name, **kwargs)
        deploy_cli = DeployServerClient(deploy_server.domain_name, deploy_auth)
        deployments = deploy_cli.list_deployments()
        return [deploy for deploy in deployments]

    @access_token()
    def delete_deployment(self, deploy_server_name: str, deployment_name: str, username: str, password: str, **kwargs):
        """Deletes model deployment

        Args:
            deploy_server_name (str): Name of deploy server
            deployment_name (str): remote model deployment name
            username (str): admin name for JupyterHub or MLflow Standalone
            password (str): admin password for JupyterHub or Mlflow Standalone

        Returns:
            None
        """
        deploy_auth = UserAuth(username=username, password=password)
        deploy_server = self._get_active_instance_by_name(instance_name=deploy_server_name, **kwargs)
        deploy_cli = DeployServerClient(deploy_server.domain_name, deploy_auth)
        deploy_cli.delete_deployment(deployment_name)

    @access_token()
    def get_deployment(self, deploy_server_name: str, deployment_name: str, username: str, password: str, **kwargs) -> DeployOut:
        """Get model deployment information

        Args:
            deploy_server_name (str): Name of deploy server
            deployment_name (str): remote model deployment name
            username (str): admin name for JupyterHub or MLflow Standalone
            password (str): admin password for JupyterHub or Mlflow Standalone

        Returns:
            DeployOut: Information about model deployments(URI, status, etc.)
        """
        deploy_auth = UserAuth(username=username, password=password)
        deploy_server = self._get_active_instance_by_name(instance_name=deploy_server_name, **kwargs)
        deploy_cli = DeployServerClient(deploy_server.domain_name, deploy_auth)
        deployment = deploy_cli.get_deployment(deployment_name)
        return deployment

    @access_token()
    def predict_model(
            self, data: Dict, deploy_server_name: str, deployment_name: str,
            model_username: str, model_password: str, **kwargs
    ) -> Dict[str, Any]:
        """Sends data for remote inference

        Args:
            data (Dict): model input data for inference
            deploy_server_name (str): Name of deploy server
            deployment_name (str): remote model deployment name
            model_username (str): username used when creating model auth e.g. "admin:password" will be "admin"
            model_password (str): password used when creating model auth e.g. "admin:password" will be "password"

        Returns:
            Dict[str, Any]: inference result, e.g. {"predictions": [161.95369430265362, 159.0451149799379]}
        """
        model_auth = UserAuth(username=model_username, password=model_password)
        deploy_server = self._get_active_instance_by_name(instance_name=deploy_server_name, **kwargs)
        model_cli = DeployModelClient(deploy_server.domain_name, deployment_name, model_auth)
        return model_cli.predict(data)

    @access_token()
    def ping_model(
            self, deploy_server_name: str, deployment_name: str, model_username: str, model_password: str, **kwargs
    ) -> str:
        """Ping deploy model

        Args:
            deploy_server_name (str): Name of deploy server
            deployment_name (str): remote model deployment name
            model_username (str): username used when creating model auth e.g. "admin:password" will be "admin"
            model_password (str): password used when creating model auth e.g. "admin:password" will be "password"

        Returns:
            (str): string containing new line symbol e.g "\n"
        """
        model_auth = UserAuth(username=model_username, password=model_password)
        deploy_server = self._get_active_instance_by_name(instance_name=deploy_server_name, **kwargs)
        model_cli = DeployModelClient(deploy_server.domain_name, deployment_name, model_auth)
        return model_cli.ping()
