#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any, Dict, Union

from pydantic import ValidationError
from requests import Session, Response, exceptions

import mlplatform_client.core.exceptions as base_exc
from mlplatform_client.core.settings import ClientSettings, APP_SETTINGS


MAX_RETRY = APP_SETTINGS.app.max_retry
CLIENT_TIMEOUT = APP_SETTINGS.app.clients_timeout


class BaseClient:
    def __init__(self, config: ClientSettings):
        self._config = config
        self.session = Session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def stop(self):
        self.session.close()

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def base_path(self) -> str:
        return self._config.base_url.strip('/')

    def get_path(self, url: str, **kwargs) -> str:
        url = url.lstrip('/')
        params = "&".join(f"{k}={v}" for k, v in kwargs.items() if v) if kwargs else None
        url = f"{url}?{params}" if params else url
        return f"{self.base_path}/{url}"

    @staticmethod
    def load_schema(data: Dict, schema):
        try:
            return schema(**data)
        except (TypeError, ValidationError) as exc:
            raise base_exc.MLPlatformException("ValidationError", f"Error while loading {schema.__name__}: {str(exc)}")

    @staticmethod
    def _raise_for_status(resp: Response):
        if resp.status_code in (200, 201, 202, 203, 204):
            return

        if resp.status_code == 404:
            raise base_exc.NotFoundException(str(resp), resp.text)
        if resp.status_code == 401:
            raise base_exc.UnauthorizedException(str(resp), resp.text)
        if resp.status_code == 403:
            raise base_exc.ForbiddenException(str(resp), resp.text)
        if resp.status_code == 409:
            raise base_exc.ConflictException(str(resp), resp.text)
        if resp.status_code == 422:
            raise base_exc.UnprocessableEntityException(str(resp), resp.text)
        if resp.status_code == 451:
            raise base_exc.UnavailableForLegalReasonException(str(resp), resp.text)
        if resp.status_code == 502:
            raise base_exc.BadGatewayException(str(resp), resp.text)
        if 400 <= resp.status_code < 500:
            raise base_exc.MLPlatformException(str(resp), resp.text)

        raise base_exc.BaseClientError(resp, resp.text)

    def _handle_response(self, resp: Response) -> Union[Dict, str, None]:
        self._raise_for_status(resp)
        if resp.status_code == 204:
            return
        
        try:
            return resp.json()
        except exceptions.JSONDecodeError:
            return resp.text

    def _perform_request(self, method: str, url: str, **kwargs) -> Any:
        with self.session.request(method, url, timeout=APP_SETTINGS.app.clients_timeout, **kwargs) as resp:
            return self._handle_response(resp)
