#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Optional, List

from mlplatform_client.clients import BaseClient
from mlplatform_client.core.exceptions import UnauthorizedException
from mlplatform_client.serializers.auth import Token, TokenRegisterIn, DeleteTokenPayload, TokenInfo
from mlplatform_client.core.settings import ClientSettings


class AuthClient(BaseClient):
    """Client for get register, refresh, and access tokens"""

    def __init__(self, config: ClientSettings, refresh_token: Optional[str] = None):
        """
        Args:
            config (ClientSettings): config for connection
            refresh_token (Optional[str]): refresh token
        """
        super().__init__(config)
        self.refresh_token: str = refresh_token
        self.access_token = None

    @staticmethod
    def headers(token: str):
        """Get headers"""
        return {"Authorization": f"Bearer {token}"}

    def _check_refresh_token(self):
        if self.refresh_token is None:
            raise UnauthorizedException(extra_info="Required refresh_token")

    def who_am_i(self) -> TokenInfo:
        """Get info about token

        Returns:
            TokenInfo: Token info
        """
        url = self.get_path("whoami")
        data = self._perform_request("get", url, headers=self.headers(self.access_token))
        return self.load_schema(data, TokenInfo)

    def get_project_tokens_info(self) -> List[TokenInfo]:
        """Get info about project tokens

        Returns:
            List[TokenInfo]: Project tokens info
        """
        url = self.get_path("all")
        data = self._perform_request("get", url, headers=self.headers(self.access_token))
        return [self.load_schema(token, TokenInfo) for token in data]

    def create_register_token(self, payload: TokenRegisterIn) -> str:
        """Creates a register token with admin or user rights, token allows to create refresh token. The register token
        for other admins and users is allowed to be created only by users with admin rights.

        Args:
            payload (TokenRegisterIn): contains client_name, token_type, access_ttl, refresh_ttl

        Returns:
            str: register token
        """
        url = self.get_path("register")
        data = self._perform_request(
            "post", url, json=payload.not_null_values(), headers=self.headers(self.access_token)
        )
        res: Token = self.load_schema(data, Token)
        return res.token

    def create_refresh_token(self, register_token: str) -> str:
        """Creates a refresh token, needed to perform all actions on the MLPlatform

        Args:
            register_token (str): register token

        Returns:
            str: refresh token
        """
        url = self.get_path("refresh")
        data = self._perform_request("post", url, headers=self.headers(register_token))
        res: Token = self.load_schema(data, Token)
        self.refresh_token = res.token
        return self.refresh_token

    def create_access_token(self) -> str:
        """Creates an access token, needed to perform all actions on the MLPlatform

        Returns:
            str: access token
        """
        self._check_refresh_token()
        url = self.get_path("access")
        data = self._perform_request("post", url, headers=self.headers(self.refresh_token))
        res: Token = self.load_schema(data, Token)
        self.access_token = res.token
        return self.access_token

    def delete_token(self, req: DeleteTokenPayload) -> None:
        """Delete token by parameters.

        Args:
            req (DeleteTokenPayload): Deletion parameters
        """
        self._perform_request('delete', f"{self.base_path}?{req.query}", headers=self.headers(self.access_token))
