import bcrypt
from typing import List

from mlplatform_client.clients import BaseClient
from mlplatform_client.serializers.settings import ClientSettings
from mlplatform_client.serializers.deployment import DeployOut, DeploymentsList, UserAuth


class DeployServerClient(BaseClient):
    """ Client for Deploy Server to deploy models for remote inference"""

    def __init__(self, deploy_server_domain: str, user_auth: UserAuth):
        super().__init__(ClientSettings(host=f"https://{deploy_server_domain}"))
        self.__auth_str = user_auth.get_basicauth()

    def auth_header(self):
        """Get headers"""
        return {"Authorization": f"Basic {self.__auth_str}"}

    def _prepare_basic_auth(self, user_auth: UserAuth) -> str:
        pass_hash = bcrypt.hashpw(user_auth.password.encode(), bcrypt.gensalt())
        return f"{user_auth.username}:{pass_hash.decode('utf-8')}"

    @staticmethod
    def deployment_exist(deploy_list: List[DeployOut], deployment_name: str) -> bool:
        filtered_list = [deploy for deploy in deploy_list if deploy.name == deployment_name]
        return True if filtered_list else False

    @staticmethod
    def find_next_port(deploy_list: List[DeployOut], port_out: int = None) -> int:
        min_port = 62000
        max_port = 65000
        deploy_server_port = 62221
        port_range = f"[{min_port} - {max_port}]"
        already_used_ports = [deploy.get_out_port() for deploy in deploy_list]

        if port_out is not None:
            if port_out < min_port or port_out > max_port:
                raise ValueError(f"Port must be in range: {port_range}")
            if port_out == deploy_server_port:
                raise ValueError(f"Port {port_out} not available. Please choose another one in range: {port_range}")
            if port_out in already_used_ports:
                raise ValueError(
                    f"Port {port_out} already in used. Please choose another one in range: {port_range}")
        else:
            for port in range(min_port, max_port):
                if port != deploy_server_port and port not in already_used_ports:
                    port_out = port
                    break
            if not port_out:
                raise ValueError(f"Deploy server has no any available ports in range: {port_range}")

        return port_out

    def create_deployment(
            self, deployment_name: str, model_uri: str, port_out: int, model_user_auth: UserAuth
    ) -> DeployOut:
        """Creates model deployment for remote inference

        Args:
            deployment_name (str): remote model deployment name
            model_uri (str): MLFlow model URI from artifact storage
            port_out (str): Remote model port
            model_user_auth (UserAuth): Model containing username and password for inference server

        Returns:
            DeployOut: Information about model deployments(URI, status, etc.)

        """
        payload = {
            "model_name": deployment_name,
            "model_uri": model_uri,
            "port_out": port_out,
            "auth": self._prepare_basic_auth(model_user_auth),
        }
        data = self._perform_request("post", self.get_path("deployments"), json=payload, headers=self.auth_header())
        return self.load_schema(data, DeployOut)

    def list_deployments(self) -> List[DeployOut]:
        """Lists all available model deployments

        Returns:
            List[DeployOut]: List of model deployments information(URI, status, etc.)
        """
        data = self._perform_request("get", self.get_path("deployments"), headers=self.auth_header())
        resp: DeploymentsList = self.load_schema(data, DeploymentsList)
        return resp.deployments

    def get_deployment(self, deployment_name: str) -> DeployOut:
        """Get model deployment information

        Args:
            deployment_name (str): model deployment name

        Returns:
            DeployOut: Information about model deployments(URI, status, etc.)
        """
        data = self._perform_request("get", self.get_path(f"deployments/{deployment_name}"), headers=self.auth_header())
        return self.load_schema(data, DeployOut)

    def delete_deployment(self, deployment_name: str) -> None:
        """Deletes model deployment

        Args:
            deployment_name (str): model deployment name

        Returns:
            None
        """
        self._perform_request("delete", self.get_path(f"deployments/{deployment_name}"), headers=self.auth_header())
