#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List

from mlplatform_client.core.exceptions import NotFoundException
from mlplatform_client.clients import BaseClient
from mlplatform_client.serializers.nova import NovaFlavor
from mlplatform_client.serializers.gateway import (
    VmDetailInfo,
    MLPlatformIN,
    BucketIn,
    NetworkLocation,
    NetworkOut,
)


class GatewayClient(BaseClient):
    """Client for actions with gateway"""

    @staticmethod
    def headers(token: str):
        """Get headers"""
        return {"Authorization": f"Bearer {token}"}

    def wake_up_cluster(self, access_token: str, cluster_id: str) -> None:
        """Activate k8s cluster

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)

        Returns:
            None
        """
        url = self.get_path(f"{cluster_id}/activate")
        return self._perform_request("post", url, headers=self.headers(access_token))

    def get_flavors(self, access_token: str) -> List[NovaFlavor]:
        """[ADMIN] Get flavors

        Args:
            access_token (str): access token

        Returns:
            List[NovaFlavor]: list with flavors id
        """
        url = self.get_path("flavors")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return [self.load_schema(flavor, NovaFlavor) for flavor in data]

    def get_instance_info(self, access_token: str, instance_id: str) -> VmDetailInfo:
        """[ADMIN] Get instance info (JupiterHub/MLFlow/Deploy)

        Args:
            access_token (str): access token
            instance_id (str): id of instance (JupiterHub/MLFlow/Deploy)

        Returns:
            VmDetailInfo: instance info (JupiterHub/MLFlow/Deploy)
        """
        url = self.get_path(f"instances/{instance_id}")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return self.load_schema(data, VmDetailInfo)

    def get_all_instances_info(self, access_token: str) -> List[VmDetailInfo]:
        """[ADMIN] Get all instances (JupiterHub/MLFlow/Deploy)

        Args:
            access_token (str): access token

        Returns:
            List[VmDetailInfo]: list with instances info (JupiterHub/MLFlow/Deploy)
        """
        url = self.get_path("instances")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        instances_info = [self.load_schema(info, VmDetailInfo) for info in data]
        return instances_info

    def get_instance_by_name(self, access_token: str, instance_name: str) -> VmDetailInfo:
        """Get instance by name (JupyterHub/MLFlow/Deploy)

        Args:
            access_token (str): access token
            instance_name (str): instance name

        Returns:
            VmDetailInfo: instance info (JupiterHub/MLFlow/Deploy)
        """
        instances = self.get_all_instances_info(access_token)
        filtered = [instance for instance in instances if instance.name == instance_name]
        if not filtered:
            raise NotFoundException(f"Instance with name: '{instance_name}' not found")
        return filtered[0]


    def create_instance(self, access_token: str, payload: MLPlatformIN) -> VmDetailInfo:
        """[ADMIN] Create JupiterHub/MLFlow/Deploy

        Args:
            access_token (str): access token
            payload (MLPlatformIN): payload for creating JupiterHub/MLFlow/Deploy

        Returns:
            VmDetailInfo: detail info about JupiterHub/MLFlow/Deploy
        """
        url = self.get_path("instances")
        data = self._perform_request(
            "post", url, json=payload.not_null_values(), headers=self.headers(access_token)
        )
        return self.load_schema(data, VmDetailInfo)

    def delete_instance(self, access_token: str, instance_id: str) -> None:
        """[ADMIN] Delete instance by id (JupiterHub/MLFlow/Deploy)

        Args:
            access_token (str): access token
            instance_id (str): id of instance (JupiterHub/MLFlow/Deploy)

        Returns:
            None
        """
        url = self.get_path(f"instances/{instance_id}")
        return self._perform_request("delete", url, headers=self.headers(access_token))

    def attach_s3(self, access_token: str, jh_id: str, payload: BucketIn):
        """[ADMIN] Attach S3 bucket Jupiter Hubs by id and bucket name

        Args:
            access_token (str): access token
            jh_id (str): id of instance Jupiter Hub
            payload (BucketIn): name of s3 bucket

        Returns:
            None
        """
        url = self.get_path(f"/instances/{jh_id}/bucket")
        return self._perform_request(
            "post", url, json=payload.not_null_values(), headers=self.headers(access_token)
        )

    def get_internal_networks(self, access_token: str) -> List[NetworkOut]:
        """[ADMIN] Get all internal networks info

        Args:
            access_token (str): access token

        Returns:
            List[NetworkOut]: list with internal networks info
        """
        url = self.get_path(f"networks?location={NetworkLocation.internal.value}")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        networks_info = [self.load_schema({**info, "external": False}, NetworkOut) for info in data]
        return networks_info

    def get_external_networks(self, access_token: str) -> List[NetworkOut]:
        """[ADMIN] Get all external networks info

        Args:
            access_token (str): access token

        Returns:
            List[NetworkOut]: list with external networks info
        """
        url = self.get_path(f"networks?location={NetworkLocation.external.value}")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        networks_info = [self.load_schema({**info, "external": True}, NetworkOut) for info in data]
        return networks_info
