#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging

from typing import List, Optional

from mlplatform_client.clients import BaseClient
from mlplatform_client.serializers.spark_proxy import (
    K8sSparkJobInfo,
    K8sSparkJobInfoDetail,
    K8sSparkJobLogs,
    K8sSparkEvent,
    K8sSparkJobsInfoList,
    K8sSparkJobIn,
    K8sClusterInfo,
    K8sSparkEventsList,
    K8sSparkApplication,
    K8sSecret,
    K8sConfigmap,
    K8sManifestIn,
    K8sObjectNamesList,
    K8sInactiveTimeoutIn,
)

log = logging.getLogger()


class SparkClient(BaseClient):
    """Client for actions with spark cluster"""

    @staticmethod
    def headers(token: str):
        """Get headers"""
        return {"Authorization": f"Bearer {token}"}

    def get_clusters(self, access_token: str) -> List[K8sClusterInfo]:
        """Get all available k8s clusters

        Args:
            access_token (str): access token

        Returns:
            List[K8sClusterInfo]: list with k8s clusters info
        """
        url = self.get_path(f"clusters")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return [self.load_schema(cluster, K8sClusterInfo) for cluster in data]

    def get_cluster_info(self, cluster_id: str, access_token: str) -> K8sClusterInfo:
        """Get info about k8s cluster

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            access_token (str): access token

        Returns:
            K8sClusterInfo: info about k8s cluster
        """
        url = self.get_path(f"clusters/{cluster_id}")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return self.load_schema(data, K8sClusterInfo)

    def set_cluster_timeout(self, access_token: str, cluster_id: str, payload: K8sInactiveTimeoutIn) -> K8sClusterInfo:
        """Sets cluster suspend or delete timeout

        Args:
            cluster_id (str): cluster id (can get from get_clusters method)
            access_token (str): access token
            payload (K8sInactiveTimeoutIn): payoad containing type of timeout to set and timeout time in minutes

        Returns:
            K8sClusterInfo: info about k8s cluster
        """
        url = self.get_path(f"clusters/{cluster_id}/set-timeout")
        data = self._perform_request("patch", url, json=payload.not_null_values(), headers=self.headers(access_token))
        return self.load_schema(data, K8sClusterInfo)

    def submit_job(self, access_token: str, cluster_id: str, payload: K8sSparkJobIn) -> K8sSparkJobInfo:
        """Submit spark job to k8s cluster

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            payload (K8sSparkJobIn): payload sent to k8s service for applying spark job

        Returns:
            K8sSparkJobInfo: info about spark job
        """
        url = self.get_path(f"/{cluster_id}/jobs")
        data = self._perform_request("post", url, json=payload.not_null_values(), headers=self.headers(access_token))
        return self.load_schema(data, K8sSparkJobInfo)

    def jobs_list(self, access_token: str, cluster_id: str) -> K8sSparkJobsInfoList:
        """Get list of spark jobs info

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)

        Returns:
            K8sSparkJobsInfoList: list with info about spark jobs
        """
        url = self.get_path(f"/{cluster_id}/jobs")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        jobs_info = [self.load_schema(info, K8sSparkJobInfo) for info in data]
        return K8sSparkJobsInfoList(jobs_info)

    def job_info(self, access_token: str, cluster_id: str, job_name: str) -> K8sSparkJobInfoDetail:
        """Get detail info of spark job

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)

        Returns:
            K8sSparkJobInfoDetail: detail info about spark job
        """
        url = self.get_path(f"{cluster_id}/jobs/{job_name}")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return self.load_schema(data, K8sSparkJobInfoDetail)

    def job_logs(self, access_token: str, cluster_id: str, job_name: str) -> K8sSparkJobLogs:
        """Get logs of spark job

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)

        Returns:
            K8sSparkJobLogs: logs of spark job
        """
        url = self.get_path(f"{cluster_id}/jobs/{job_name}/logs")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return self.load_schema(data, K8sSparkJobLogs)

    def describe_job(self, access_token: str, cluster_id: str, job_name: str) -> K8sSparkApplication:
        """Describes job by its name

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)

        Returns:
            K8sSparkApplication: description of spark job (application)
        """
        url = self.get_path(f"{cluster_id}/jobs/{job_name}/describe")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return self.load_schema(data, K8sSparkApplication)

    def events(self, access_token: str, cluster_id: str) -> K8sSparkEventsList:
        """Get events of spark cluster

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)

        Returns:
            K8sSparkEventsList: list of spark cluster events
        """
        url = self.get_path(f"{cluster_id}/events")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        events = []
        for event_data in data:
            try:
                event = self.load_schema(event_data, K8sSparkEvent)
                events.append(event)
            except Exception as exc:
                log.warning(exc)
        return K8sSparkEventsList(events)

    def delete_job(self, access_token: str, cluster_id: str, job_name: str):
        """Delete spark job

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)
        """
        url = self.get_path(f"{cluster_id}/jobs/{job_name}")
        return self._perform_request("delete", url, headers=self.headers(access_token))
    
    def get_default_manifest(self, access_token: str, cluster_id: str, job_name: Optional[str] = None) -> str:
        """Get SparkApplication default manifest

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            job_name (str): job name from the yaml manifest (metadata.name)

        Returns:
            str: get example of SparkApplication default manifest
        """
        url = self.get_path(f"{cluster_id}/manifest", job_name=job_name)
        return self._perform_request("get", url, headers=self.headers(access_token))

    def create_secret(
            self, access_token: str, cluster_id: str, payload: K8sManifestIn, namespace: str = "default"
    ) -> K8sSecret:
        """Create k8s secret

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            payload (K8sManifestIn): payload for applying secret manifest on remote k8s cluster
            namespace (str): namespace of a secret

        Returns:
            K8sSecret: detail info about k8s secret
        """
        url = self.get_path(f"{cluster_id}/secrets?namespace={namespace}")
        data = self._perform_request("post", url, json=payload.not_null_values(), headers=self.headers(access_token))
        return self.load_schema(data, K8sSecret)

    def get_secret(
            self, access_token: str, cluster_id: str, secret_name: str, namespace: str = "default"
    ) -> K8sSecret:
        """Get k8s secret

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            secret_name (str): name of a secret on k8s cluster
            namespace (str): namespace of a secret

        Returns:
            K8sSecret: detail info about k8s secret
        """
        url = self.get_path(f"{cluster_id}/secrets/{secret_name}?namespace={namespace}")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return self.load_schema(data, K8sSecret)

    def list_secrets(self, access_token: str, cluster_id: str, namespace: str = "default") -> K8sObjectNamesList[str]:
        """List k8s secrets

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            namespace (str): namespace of a secret

        Returns:
            K8sObjectNamesList[str]: k8s secrets names
        """
        url = self.get_path(f"{cluster_id}/secrets?namespace={namespace}")
        data_list = self._perform_request("get", url, headers=self.headers(access_token))
        return K8sObjectNamesList([self.load_schema(data, K8sSecret).name for data in data_list])

    def delete_secret(self, access_token: str, cluster_id: str, secret_name: str, namespace: str = "default"):
        """Delete k8s secret

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            secret_name (str): name of a secret on k8s cluster
            namespace (str): namespace of a secret

        """
        url = self.get_path(f"{cluster_id}/secrets/{secret_name}?namespace={namespace}")
        return self._perform_request("delete", url, headers=self.headers(access_token))

    def create_configmap(
            self, access_token: str, cluster_id: str, payload: K8sManifestIn, namespace: str = "default"
    ) -> K8sConfigmap:
        """Create k8s configMap

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            payload (K8sManifestIn): payload for applying configMap manifest on remote k8s cluster
            namespace (str): namespace of a secret

        Returns:
            K8sSecret: detail info about k8s secret
        """
        url = self.get_path(f"{cluster_id}/configmaps?namespace={namespace}")
        data = self._perform_request("post", url, json=payload.not_null_values(), headers=self.headers(access_token))
        return self.load_schema(data, K8sConfigmap)

    def get_configmap(
            self, access_token: str, cluster_id: str, configmap_name: str, namespace: str = "default"
    ) -> K8sConfigmap:
        """Get k8s configMap

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            configmap_name (str): name of a configMap on k8s cluster
            namespace (str): namespace of a configMap

        Returns:
            K8sConfigmap: detail info about k8s configMap
        """
        url = self.get_path(f"{cluster_id}/configmaps/{configmap_name}?namespace={namespace}")
        data = self._perform_request("get", url, headers=self.headers(access_token))
        return self.load_schema(data, K8sConfigmap)

    def list_configmaps(self, access_token: str, cluster_id: str, namespace: str = "default") -> K8sObjectNamesList[str]:
        """List k8s configMaps

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            namespace (str): namespace of a configMap

        Returns:
            K8sObjectNamesList[str]: k8s configmap names
        """
        url = self.get_path(f"{cluster_id}/configmaps?namespace={namespace}")
        data_list = self._perform_request("get", url, headers=self.headers(access_token))
        return K8sObjectNamesList([self.load_schema(data, K8sConfigmap).name for data in data_list])

    def delete_configmap(self, access_token: str, cluster_id: str, configmap_name: str, namespace: str = "default"):
        """Delete k8s configMap

        Args:
            access_token (str): access token
            cluster_id (str): cluster id (can get from get_clusters method)
            configmap_name (str): name of a configMap on k8s cluster
            namespace (str): namespace of a configMap

        """
        url = self.get_path(f"{cluster_id}/configmaps/{configmap_name}?namespace={namespace}")
        return self._perform_request("delete", url, headers=self.headers(access_token))
