#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
import base64
from functools import wraps
from typing import Callable, List

from mlplatform_client.core.exceptions import UnauthorizedException, WrongTokenException


def access_token(throw_access_token: bool = True) -> Callable:
    """Get or create access token decorator

    Args:
        throw_access_token: Pass a parameter `access_token` into the called function

    Returns:
        Function decorator
    """
    def decorator(func: Callable) -> Callable:
        """Get or create access token

        Args:
            func: A function that takes an access token as an argument

        Returns:
            func: result of passed function
        """
        def call_func(self, *args, **kwargs):
            """Call wrapped function and throw `access_token` parameter if required"""
            if throw_access_token:
                kwargs.update(access_token=self._auth_client.access_token)
            return func(self, *args, **kwargs)

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                if not self._auth_client.access_token:
                    self._auth_client.create_access_token()
                return call_func(self, *args, **kwargs)
            except UnauthorizedException:
                self._auth_client.create_access_token()
                return call_func(self, *args, **kwargs)
        return wrapper
    return decorator


def decode_token(token: str) -> list:
    """Decodes token
       Args:
            token: Token passed by user in function or method
       Returns:
            boolean: returns list of token suffixies
    """
    t_parts = token.split('.')
    correct_token_parts_num = 3
    suffixies = []
    if len(t_parts) == correct_token_parts_num:
        raw_open_info = t_parts[1]
        prepped_open_info = f"{raw_open_info}{'=' * (4 - len(raw_open_info) % 4)}"
        bs64decoded_string = base64.b64decode(prepped_open_info).decode()
        for element in json.loads(bs64decoded_string)['aud']:
            suffixies.append(element.split('/')[-1])
    return suffixies


def get_token_names_passed_by_user(prefix_x_token_suffix: dict, token_suffixies: List[str]):
    """Gets types of tokens passed by user

    Args:
        prefix_x_token_suffix: Token prefix & suffix compliance scheme
        token_suffixies: Token suffixies retrieved from passed token

    Returns:
        Function decorator
    """
    passed_token_names = []
    suffix_x_prefix = {v: k for k, v in prefix_x_token_suffix.items()}
    if token_suffixies:
        passed_token_prefixies = [suffix_x_prefix.get(i) for i in token_suffixies]
        passed_token_names.extend([f"{n}_token" for n in passed_token_prefixies])
    else:
        passed_token_names.append("access_token")
    return passed_token_names


def validate_token(necessary_token_prefix: str = "register") -> Callable:
    """Validates that right token was passed inside method

    Args:
        necessary_token_prefix: Pass a prefix of a necessary token

    Returns:
        Function decorator
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            register_token_suffix = "CreateRefresh"
            refresh_token_suffix = "CreateAccess"
            prefix_x_token_suffix = {
                "register": register_token_suffix,
                "refresh": refresh_token_suffix
            }
            necessary_token_name = f"{necessary_token_prefix}_token"
            necessary_token_suffix = prefix_x_token_suffix.get(necessary_token_prefix)
            raw_token = kwargs.get(necessary_token_name) or args[0]
            token_suffixies = decode_token(raw_token)
            if necessary_token_suffix in token_suffixies and necessary_token_prefix != "access":
                return func(self, *args, **kwargs)
            passed_token_names = get_token_names_passed_by_user(prefix_x_token_suffix, token_suffixies)
            raise WrongTokenException("Wrong token passed, method needs: %s, passed: %s"
                                      % (necessary_token_name, ", ".join(passed_token_names)))
        return wrapper
    return decorator
