#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, List, Optional
import json
import yaml

from mlplatform_client.serializers.spark_proxy import K8sSparkJobRestartPolicyType


class Manifest:
    """Create manifest from yaml file"""
    def __init__(self, dict_raw: Dict):
        self._dict_raw = dict_raw
    
    @classmethod
    def new_from_yaml(cls, yaml_raw: str) -> "Manifest":
        """Create manifest class from raw yaml

        Args:
            yaml_raw (str): manifest in yaml format as string

        Returns:
            Manifest: instance of class Manifest
        """
        return cls(yaml.safe_load(yaml_raw))
    
    def to_yaml(self) -> str:
        """Convert dict manifest to string

        Returns:
            str: manifest as string
        """
        return yaml.safe_dump(self._dict_raw, default_flow_style=False)

    def save_yaml(self, file_path: str) -> None:
        """Save Manifest as yaml file

        Args:
            file_path (str): path to save yaml file

        Returns:
            None
        """
        with open(file_path, "w") as f:
            f.write(self.to_yaml())

    @classmethod
    def load_yaml(cls, file_path: str) -> "Manifest":
        """Convert yaml file to Manifest

        Args:
            file_path: path to yaml file

        Returns:
            Manifest: instance of class Manifest
        """
        with open(file_path, "r") as f:
            return cls.new_from_yaml(f.read())

    def __str__(self) -> str:
        return json.dumps(self._dict_raw, indent=3)
    
    @property
    def job_name(self):
        return self._dict_raw["metadata"]["name"]

    @job_name.setter
    def job_name(self, job_name: str):
        self._dict_raw["metadata"]["name"] = job_name

    @property
    def main_app_file(self):
        return self._dict_raw["spec"]["mainApplicationFile"]

    @main_app_file.setter
    def main_app_file(self, main_app_file: str):
        self._dict_raw["spec"]["mainApplicationFile"] = main_app_file

    @property
    def main_class(self):
        return self._dict_raw["spec"]["mainClass"]

    @main_class.setter
    def main_class(self, main_class: str):
        self._dict_raw["spec"]["mainClass"] = main_class

    @property
    def spec_type(self):
        return self._dict_raw["spec"]["type"]

    @spec_type.setter
    def spec_type(self, spec_type: str):
        self._dict_raw["spec"]["type"] = spec_type

    def _set_settings(self, env_type: str, spark_settings: Dict):
        """Sets spark operator manifest settings

        Args:
            env_type (str): type of settings to set.
            spark_settings (dict): dict containing manifest settings e.g. cores, env, envFrom

        Returns:
            None
        """
        settings = self._dict_raw["spec"].get(env_type, {})
        settings.update(**spark_settings)
        self._dict_raw["spec"][env_type] = settings

    def set_driver_settings(self, driver_settings: Dict):
        """Sets spark operator manifest settings for driver

        Args:
            driver_settings (dict): dict containing driver settings e.g. cores, env, envFrom

        Returns:
            None
        """
        self._set_settings("driver", driver_settings)

    def set_executor_settings(self, executor_settings: Dict):
        """Sets spark operator manifest settings for executor

        Args:
            executor_settings (dict): dict containing executor settings e.g. cores, env, envFrom

        Returns:
            None
        """
        self._set_settings("executor", executor_settings)

    def set_spark_conf(self, spark_conf: Dict):
        """Sets spark operator manifest settings for sparkConf

        Args:
            spark_conf (dict): dict containing spark configuration settings e.g. spark.eventLog.dir

        Returns:
            None
        """
        self._set_settings("sparkConf", spark_conf)

    def _set_deps(self, deps_type: str, deps_list: List[str]):
        """Sets spark operator manifest dependencies

        Args:
            deps_type (str): type of spark operator manifest dependencies e.g. jars, files, pyFiles
            deps_list (list): list containing spark dependencies e.g. ["s3a://bucket/dataset.csv"]

        Returns:
            None
        """
        deps = {deps_type: deps_list}
        self._set_settings("deps", deps)

    def set_jars(self, jars_list: List[str]):
        """Sets spark operator manifest jars dependencies

        Args:
            jars_list (list): list containing spark jars dependencies e.g. ["s3a://bucket/spark-pi.jar"]

        Returns:
            None
        """
        self._set_deps("jars", jars_list)

    def set_files(self, files_list: List[str]):
        """Sets spark operator manifest files dependencies

        Args:
            files_list (list): list containing spark files dependencies e.g. ["s3a://bucket/dataset.csv"]

        Returns:
            None
        """
        self._set_deps("files", files_list)

    def set_pyfiles(self, pyfiles_list: List[str]):
        """Sets spark operator manifest pyFiles dependencies

        Args:
            pyfiles_list (list): list containing spark pyFiles dependencies e.g. ["s3a://bucket/custom_job.py"]

        Returns:
            None
        """
        self._set_deps("pyFiles", pyfiles_list)

    def set_repositories(self, repositories_list: List[str]):
        """Sets spark operator manifest repositores dependencies

        Args:
            repositories_list (list): list containing spark repositores dependencies e.g. ["https://repository.example.com/prod"]

        Returns:
            None
        """
        self._set_deps("repositories", repositories_list)

    def set_packages(self, packages_list: List[str]):
        """Sets spark operator manifest packages dependencies

        Args:
            packages_list (list): list containing spark packages dependencies e.g. ["com.example:some-package:1.0.0"]

        Returns:
            None
        """
        self._set_deps("packages", packages_list)

    def set_restart_policy(
        self,
        restart_type: K8sSparkJobRestartPolicyType = K8sSparkJobRestartPolicyType.NEVER,
        on_failure_retries: Optional[int] = None,
        on_failure_retry_interval: Optional[int] = None,
        on_submission_failure_retries: Optional[int] = None,
        on_submission_failure_retry_interval: Optional[int] = None,
    ):
        """
        Set the restart policy for the Kubernetes Spark job.

        Args:
            restart_type (K8sSparkJobRestartPolicyType): The type of restart policy
                                                         (default is `K8sSparkJobRestartPolicyType.NEVER`).
            on_failure_retries (int): The number of retries on failure.
            on_failure_retry_interval (int): The interval between retries on failure.
            on_submission_failure_retries (int): The number of retries on submission failure.
            on_submission_failure_retry_interval (int): The interval between retries on submission failure.

        Returns:
            None
        """
        restart_policy = {
            "type": restart_type.value,
        }
        if restart_type != K8sSparkJobRestartPolicyType.NEVER:
            if on_failure_retries:
                restart_policy["onFailureRetries"] = on_failure_retries
            if on_failure_retry_interval:
                restart_policy["onFailureRetryInterval"] = on_failure_retry_interval
            if on_submission_failure_retries:
                restart_policy["onSubmissionFailureRetries"] = on_submission_failure_retries
            if on_submission_failure_retry_interval:
                restart_policy["onSubmissionFailureRetryInterval"] = on_submission_failure_retry_interval
        self._dict_raw["spec"]["restartPolicy"] = restart_policy

    def _add_deps(self, deps_type: str, deps_list: List[str]):
        """Extends spark operator manifest jars dependencies

        Args:
            deps_type (str): Type of dependencies e.g. jars, files, pyFiles
            deps_list (list): list containing spark jars dependencies e.g. ["s3a://bucket/spark-pi.jar"]

        Returns:
            None
        """
        deps = self._dict_raw["spec"].get("deps", {})
        if not deps.get(deps_type):
            deps[deps_type] = list()
        deps[deps_type].extend(deps_list)
        self._dict_raw["spec"]["deps"] = deps

    def add_jars(self, jars_list: List[str]):
        """Extends spark operator manifest jars dependencies

        Args:
            jars_list (list): list containing spark jars dependencies e.g. ["s3a://bucket/spark-pi.jar"]

        Returns:
            None
        """
        self._add_deps("jars", jars_list)

    def add_files(self, files_list: List[str]):
        """Extends spark operator manifest files dependencies

        Args:
            files_list (list): list containing spark files dependencies e.g. ["s3a://bucket/dataset.csv"]

        Returns:
            None
        """
        self._add_deps("files", files_list)

    def add_pyfiles(self, pyfiles_list: List[str]):
        """Extends spark operator manifest pyFiles dependencies

        Args:
            pyfiles_list (list): list containing spark pyFiles dependencies e.g. ["s3a://bucket/custom_job.py"]

        Returns:
            None
        """
        self._add_deps("pyFiles", pyfiles_list)

    def add_repositories(self, repositories_list: List[str]):
        """Extends spark operator manifest repositores dependencies

        Args:
            repositories_list (list): list containing spark repositores dependencies e.g. ["https://repository.example.com/prod"]

        Returns:
            None
        """
        self._add_deps("repositories", repositories_list)

    def add_packages(self, packages_list: List[str]):
        """Extends spark operator manifest packages dependencies

        Args:
            packages_list (list): list containing spark packages dependencies e.g. ["com.example:some-package:1.0.0"]

        Returns:
            None
        """
        self._add_deps("packages", packages_list)

    def _remove_deps(self, deps_type: str):
        """Removes spark operator manifest jars dependencies

        Args:
            deps_type (str): Type of dependencies e.g. jars, files, pyFiles

        Returns:
            None
        """
        self._dict_raw["spec"].get("deps", {}).pop(deps_type, None)
        if not self._dict_raw["spec"].get("deps"):
            self._dict_raw["spec"].pop("deps", None)

    def remove_jars(self):
        """Removes spark operator manifest jars dependencies

        Returns:
            None
        """
        self._remove_deps("jars")

    def remove_files(self):
        """Removes spark operator manifest files dependencies

        Returns:
            None
        """
        self._remove_deps("files")

    def remove_pyfiles(self):
        """Removes spark operator manifest pyFiles dependencies

        Returns:
            None
        """
        self._remove_deps("pyFiles")

    def remove_repositories(self):
        """Removes spark operator manifest repositories dependencies

        Returns:
            None
        """
        self._remove_deps("repositories")

    def remove_packages(self):
        """Removes spark operator manifest packages dependencies

        Returns:
            None
        """
        self._remove_deps("packages")

    def _add_env(self, env_entity: str, env_type: str, env_list: List[str]):
        """Adds entries to env or envFrom field in spark operator manifest

        Args:
            env_entity (str): Type of env entity e.g. executor, driver
            env_type (str): Type of env field e.g. env or envFrom
            env_list (list): list containing manifest env or envFrom entries e.g. [{"name": "test", "value": "test"}]

        Returns:
            None
        """
        env = self._dict_raw["spec"].get(env_entity, {}).get(env_type, [])
        if not env:
            self._dict_raw["spec"][env_entity][env_type] = list()
        env.extend(env_list)
        self._dict_raw["spec"][env_entity][env_type] = env

    def add_executor_env(self, env_list: List[str]):
        """Adds entries to executor env

        Args:
            env_list (list): list containing manifest env entries e.g. [{"name": "test-name", "value": "test"}]

        Returns:
            None
        """
        self._add_env("executor", "env", env_list)

    def add_executor_env_from(self, env_list: List[str]):
        """Adds entries to executor envFrom

        Args:
            env_list (list): list containing manifest env entries e.g. [{"secretRef": {"name": "test-name"}}]

        Returns:
            None
        """
        self._add_env("executor", "envFrom", env_list)

    def add_driver_env(self, env_list: List[str]):
        """Adds entries to driver env

        Args:
            env_list (list): list containing manifest env entries e.g. [{"name": "test", "value": "test"}]

        Returns:
            None
        """
        self._add_env("driver", "env", env_list)

    def add_driver_env_from(self, env_list: List[str]):
        """Adds entries to driver envFrom

        Args:
            env_list (list): list containing manifest env entries e.g. [{"secretRef": {"name": "test-name"}}]

        Returns:
            None
        """
        self._add_env("driver", "envFrom", env_list)
