#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import field_validator
from pydantic_core.core_schema import ValidationInfo

from mlplatform_client.serializers import BaseSerializerModel


class MLPTokenType(str, Enum):
    ADMIN = "ADMIN"
    USER = "USER"


class Token(BaseSerializerModel):
    token: str


class TokenInfo(BaseSerializerModel):
    id: str
    name: str
    type: Optional[MLPTokenType] = None
    issuer_name: Optional[str] = None
    exp: Optional[datetime] = None

    def __str__(self):
        return f"\nToken {self.name}:" \
               f"\n\tid: {self.id}" \
               f"\n\ttype: {self.type}" \
               f"\n\tissuer_name: {self.issuer_name}" \
               f"\n\texp: {self.exp}"


class TokenRegisterIn(BaseSerializerModel):
    client_name: str
    token_type: MLPTokenType
    access_ttl: Optional[str] = None
    refresh_ttl: Optional[str] = None


class DeleteTokenPayload(BaseSerializerModel):
    token_id: Optional[str] = None
    token_name: Optional[str] = None

    @field_validator("token_id", "token_name")
    @classmethod
    def validate_all(cls, v, validation_info: ValidationInfo):
        check_list = [*validation_info.data.values(), v]
        if len(check_list) == 2 and all([value is None for value in check_list]):
            raise ValueError("One of the 'token_id' or 'token_name' parameters must be present")
        return v

    @property
    def query(self) -> str:
        query = []
        if self.token_id is not None:
            query.append(f"token_id={self.token_id}")
        if self.token_name is not None:
            query.append(f"token_name={self.token_name}")

        return '&'.join(query)
