import base64
from typing import Optional, List
from pydantic import Field, field_validator
from string import punctuation

from mlplatform_client.serializers import BaseSerializerModel


class DeployPort(BaseSerializerModel):
    host_ip: str = Field(alias="HostIp")
    host_port: str = Field(alias="HostPort")


class DeployPorts(BaseSerializerModel):
    out_ports: List[DeployPort] = Field(alias="8080/tcp")

    def collect_ports(self):
        return [item.host_port for item in self.out_ports]


class DeployOut(BaseSerializerModel):
    name: str
    status: str
    ports: DeployPorts
    image_tags: Optional[List[str]]
    model_ref: Optional[str]

    def get_out_port(self) -> int:
        ports: List[str] = self.ports.collect_ports()
        uniq_ports = set(ports)
        if len(uniq_ports) == 1:
            return int(uniq_ports.pop())
        raise ValueError(f"Deploy has multiple open ports")

    def dict(self, *args, **kwargs):
        return {
            "name": self.name,
            "status": self.status,
            "port": self.get_out_port(),
            "image_tags": self.image_tags,
            "model_ref": self.model_ref,
            "predict_ref": f"{self.model_ref}/invocations" if self.model_ref else "",
        }

    def __str__(self):
        items = ''.join("\n%s: %s" % item for item in self.dict().items())
        return f'\n---------------\nDeployment Info:\n---------------{items}'


class DeploymentsList(BaseSerializerModel):
    deployments: List[DeployOut]


class UserAuth(BaseSerializerModel):
    username: str
    password: str

    @field_validator("password")
    @classmethod
    def validate_password(cls, value):
        default_password_length = 8
        has_good_length = len(value) >= default_password_length
        if not has_good_length:
            raise ValueError(
                f"Password length must be more or equal {default_password_length}."
            )
        has_uppercase_chars = any(char.isupper() for char in value)
        if not has_uppercase_chars:
            raise ValueError("Password has no uppercase letters")

        has_lowercase_chars = any(char.islower() for char in value)
        if not has_lowercase_chars:
            raise ValueError("Password has no lowercase letters")

        has_special_chars = any(char for char in value if char in punctuation)
        if not has_special_chars:
            raise ValueError("Password has no special chars")
        return value

    def get_basicauth(self) -> str:
        return base64.standard_b64encode(f"{self.username}:{self.password}".encode()).decode('utf-8')
