import datetime
from typing import List, Optional, Tuple
from string import punctuation
from enum import Enum

from pydantic import Field, model_validator, field_validator

from mlplatform_client.serializers import BaseAuditModel, BaseSerializerModel
from mlplatform_client.core.settings import APP_SETTINGS
from mlplatform_client.core.utils import dns_slugs_validator, s3bucket_name_validator


class FrontInstanceStatus(str, Enum):
    PREPARE_DBAAS = "PREPARE_DBAAS"
    CREATING = "CREATING"
    CREATE_FAILED = "CREATE_FAILED"
    INSTALL_SCRIPTS = "INSTALL_SCRIPTS"
    RUNNING = "RUNNING"
    DELETING = "DELETING"
    DELETED = "DELETED"
    QUOTA_LIMIT = "QUOTA_LIMIT"
    LOST_CONNECTION = "LOST_CONNECTION"
    STARTING = "STARTING"
    FREEZING = "FREEZING"
    FREEZED = "FREEZED"
    BACKUP = "BACKUP"
    SNAPSHOT = "SNAPSHOT"
    BACKUP_CREATE_FAILED = "BACKUP_CREATE_FAILED"
    BACKUP_CREATE_DETACH = "BACKUP_CREATE_DETACH"
    BACKUP_CREATE_ATTACH = "BACKUP_CREATE_ATTACH"
    BACKUP_VOLUME = "BACKUP_VOLUME"
    BACKUP_RECOVER_DETACH = "BACKUP_RECOVER_DETACH"
    BACKUP_RECOVER_REVITALIZING = "BACKUP_RECOVER_REVITALIZING"
    BACKUP_RECOVER_ATTACH = "BACKUP_RECOVER_ATTACH"
    BACKUP_RECOVER_CHECK = "BACKUP_RECOVER_CHECK"
    BACKUP_RECOVER_FAILED = "BACKUP_RECOVER_FAILED"

    def __str__(self):
        return self.value


class AvailabilityZone(str, Enum):
    GZ1 = "GZ1"
    MS1 = "MS1"

    def __str__(self):
        return self.value


class VolumeType(str, Enum):
    ceph_ssd = "ceph-ssd"
    high_iops = "high-iops"

    def __str__(self):
        return self.value


class MLPlatformVolumeIn(BaseSerializerModel):
    name: Optional[str] = None
    size: int
    volume_type: VolumeType
    imageRef: Optional[str] = None
    availability_zone: AvailabilityZone
    cinder_id: Optional[str] = None

    def __str__(self):
        attrs = vars(self)
        items = ''.join("\n    %s: %s" % items for items in attrs.items())
        return f'\n---------------\n{items[1:]}'


class VmInfo(BaseAuditModel):
    id: str
    name: str
    status: FrontInstanceStatus
    flavor_id: str
    created_dt: str
    public_ip: Optional[str] = None
    private_ip: Optional[str] = None
    domain_name: Optional[str] = None
    status_reason: Optional[str] = None
    instance_type: str
    mlflow_jh_instance_id: Optional[str] = None
    mlflow_deploy_instance_id: Optional[str] = None


class VmDetailInfo(VmInfo):
    jh_admin_name: str
    availability_zone: str
    volumes: List[MLPlatformVolumeIn]

    def __str__(self):
        attrs = vars(self)
        items = ''.join("\n%s: %s" % item for item in attrs.items())
        return f'\n---------------\nJupiter Hub Detail Info:\n---------------{items}'


class InstanceType(str, Enum):
    JUPYTERHUB = "JUPYTERHUB"
    MLFLOW = "MLFLOW"
    MLFLOW_STANDALONE = "MLFLOW_STANDALONE"
    DEPLOY = "DEPLOY"

    def __str__(self):
        return self.value

    @classmethod
    def default_types(cls):
        return [
            cls.JUPYTERHUB.value,
            cls.MLFLOW.value,
            cls.DEPLOY.value,
        ]


class MLPlatformNetworkIn(BaseSerializerModel):
    network_id: str


class SDNTypes(str, Enum):
    neutron = "neutron"
    sprut = "sprut"

    def __str__(self):
        return self.value

    @classmethod
    def available_types(cls) -> Tuple[str, ...]:
        return (
            cls.neutron.value, cls.sprut.value
        )


class MLPlatformIN(BaseSerializerModel):
    instance_name: str
    domain_name: Optional[str] = Field(default=None, min_length=2, max_length=63)
    instance_type: Optional[InstanceType] = InstanceType.JUPYTERHUB
    jh_admin_name: Optional[str] = "jh_default_admin"
    jh_admin_password: Optional[str] = None
    is_gpu: Optional[bool] = False
    flavor: str
    volumes: List[MLPlatformVolumeIn]
    networks: Optional[MLPlatformNetworkIn] = None
    keypair: Optional[str] = None
    mlflow_jh_instance_id: Optional[str] = None
    deploy_mlflow_instance_id: Optional[str] = None
    is_mlflow_demo_mode: Optional[bool] = True
    base_image_id: Optional[str] = None
    scripts_for_install: Optional[List[str]] = None
    auth_token: Optional[str] = None
    s3fs_bucket: Optional[str] = Field(default=None, min_length=4, max_length=63)
    sdn: str = SDNTypes.neutron.value

    @model_validator(mode="before")
    @classmethod
    def validate_root_cert_id(cls, values):
        if not values.get("domain_name"):
            values["domain_name"] = None

        domain_name, instance_type = values["domain_name"], values.get("instance_type")

        if domain_name is None and instance_type not in [InstanceType.MLFLOW, InstanceType.DEPLOY]:
            raise ValueError(
                f"Add domain name parameter for Instance with type {instance_type}"
            )

        return values

    @staticmethod
    def _validate_domain_name(domain_name: str) -> str:
        if dns_slugs_validator.match(domain_name) is None:
            raise ValueError(f'''Incorrect '{domain_name}' domain name format. Allow just lowercase Latin letters or numbers. 
            Domain name can't start and end with '-' or '.' ''')
        return domain_name

    @field_validator('domain_name')
    @classmethod
    def validate_domain_name(cls, v: str) -> str:
        if not v:
            return
        return cls._validate_domain_name(v)

    @staticmethod
    def _validate_s3fs_bucket(bucket_name: str) -> str:
        if s3bucket_name_validator.match(bucket_name) is None:
            raise ValueError(f'''Incorrect '{bucket_name}' S3 bucket name format. Allow just lowercase Latin letters or numbers. 
            S3 bucket name can't start and end with '-' or '.' or '_' ''')
        return bucket_name

    @field_validator('s3fs_bucket')
    @classmethod
    def validate_s3fs_bucket(cls, v: str) -> str:
        if not v:
            return
        return cls._validate_s3fs_bucket(v)

    @property
    def availability_zone(self) -> str:
        av_zones = set(v.availability_zone for v in self.volumes)
        if len(av_zones) == 1:
            return str(av_zones.pop())
        raise ValueError(f"Payload has inconsistent availability_zones: {av_zones}")

    @field_validator("jh_admin_name")
    @classmethod
    def validate_jh_admin_name(cls, v):
        default_admin_name_length = 50
        if len(v) > default_admin_name_length:
            raise ValueError("Too many symbols for JUPYTER HUB administrator name")
        return v

    @field_validator("volumes")
    @classmethod
    def check_volumes_quantity(cls, v):
        if len(v) > APP_SETTINGS.app.upper_disks_number_threshold:
            raise ValueError("Too many disks in request")
        return v

    @field_validator("jh_admin_password")
    @classmethod
    def validate_jh_admin_password(cls, v):
        if not v:
            return
        default_password_length = 8
        restricted_chars = r"\/`'"
        has_good_length = len(v) >= default_password_length
        if not has_good_length:
            raise ValueError(
                f"Password length must be more or equal {default_password_length}."
            )
        has_uppercase_chars = any(char.isupper() for char in v)
        if not has_uppercase_chars:
            raise ValueError("Password has no uppercase letters")

        has_lowercase_chars = any(char.islower() for char in v)
        if not has_lowercase_chars:
            raise ValueError("Password has no lowercase letters")

        has_special_chars = any(char for char in v if char in punctuation)
        if not has_special_chars:
            raise ValueError("Password has no special chars")

        has_restricted_char = any(char for char in v if char in restricted_chars)
        if has_restricted_char:
            raise ValueError("Password has restricted characters ({restricted})".format(restricted=restricted_chars))
        return v

    @field_validator("sdn")
    @classmethod
    def check_sdn_type(cls, v):
        if v not in SDNTypes.available_types():
            raise ValueError("Unknown SDN type %s" % v)
        return v


class BucketIn(BaseSerializerModel):
    bucket_name: str = Field(min_length=4, max_length=63)

    @staticmethod
    def _validate_bucket_name(bucket_name: str) -> str:
        if s3bucket_name_validator.match(bucket_name) is None:
            raise ValueError(f'''Incorrect '{bucket_name}' S3 bucket name format. Allow just lowercase Latin letters or numbers. 
            S3 bucket name can't start and end with '-' or '.' or '_' ''')
        return bucket_name

    @field_validator('bucket_name')
    @classmethod
    def validate_bucket_name(cls, v: str) -> str:
        if not v:
            return
        return cls._validate_bucket_name(v)


class NetworkOut(BaseSerializerModel):
    id: str
    name: str
    admin_state_up: bool
    created_at: datetime.datetime
    updated_at: Optional[datetime.datetime] = None
    sdn: Optional[str] = SDNTypes.neutron.value
    external: Optional[bool] = None

    def __str__(self):
        attrs = vars(self)
        items = ''.join("\n%s: %s" % item for item in attrs.items())
        return f'\n---------------\nNetwork Info:\n---------------{items}'


class NetworkLocation(str, Enum):
    internal = "internal"
    external = "external"
