#  Copyright (c) 2023 VK Group
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#     http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from enum import Enum
from typing import Optional, List, Dict, Any
from yaml import safe_dump
import datetime

from dateutil.tz import tzutc  # very important don't delete
from pydantic import field_validator, model_validator, GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

from mlplatform_client.serializers import BaseSerializerModel


class PrettyList(list):
    def __init__(self, *args, **kwargs):
        super(PrettyList, self).__init__(args[0])

    @classmethod
    def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(list))

    def __str__(self):
        return "\n".join([str(elem) for elem in self])

    def __repr__(self):
        return self.__str__()


class K8sSparkJobStatus(str, Enum):
    ERROR = "ERROR"
    SUBMITTED = "SUBMITTED"
    DELETED = "DELETED"
    WAITING = "WAITING"
    COMPLETED = "COMPLETED"



class K8sSparkJobRestartPolicyType(str, Enum):
    NEVER = "Never"
    ON_FAILURE = "OnFailure"
    ALWAYS = "Always"


class K8sSparkJobIn(BaseSerializerModel):
    spark_manifest: str
    python_file_name: Optional[str] = None
    python_code: Optional[str] = None


class K8sSparkJobInfo(BaseSerializerModel):
    id: str
    k8s_cluster_id: str
    job_name: str
    image_pull_secret: str
    configmap_name: str
    spark_manifest: str
    status: K8sSparkJobStatus
    created_at: str
    project_id: str
    ui_url: Optional[str] = None

    def __str__(self):
        return f"job_name: {self.job_name}, status: {self.status}, created_at: {self.created_at} ui_url: {self.ui_url}"


class K8sSparkJobsInfoList(PrettyList):
    def __init__(self, *args, **kwargs):
        super(K8sSparkJobsInfoList, self).__init__(args[0])


class K8sSparkJobInfoDetail(K8sSparkJobInfo):
    pod_status: Optional[str] = None
    pod_status_reason: Optional[str] = None
    pod_state: Optional[str] = None
    pod_state_reason: Optional[str] = None
    pod_state_reason_message: Optional[str] = None

    def __str__(self):
        r = "\r"
        return f'''job_name: {self.job_name} 
status: {self.status}
created_at: {self.created_at}
ui_url: {self.ui_url}
pod_status: {getattr(self, "pod_status")}
pod_status_reason: {getattr(self, "pod_status_reason")}
pod_state: {getattr(self, "pod_state")}
pod_state_reason: {getattr(self, "pod_state_reason")}
pod_state_reason_message: {getattr(self, "pod_state_reason_message")}
        '''

    @property
    def full_info(self):
        attrs = vars(self)
        return '\n'.join("%s: %s" % item for item in attrs.items())


class MLPlatformVolumeIn(BaseSerializerModel):
    name: Optional[str] = None
    size: int
    volume_type: str
    imageRef: Optional[str] = None
    availability_zone: str
    cinder_id: Optional[str] = None

    def __str__(self):
        attrs = vars(self)
        return ''.join("\n        %s: %s" % item for item in attrs.items())


class MLPlatformVolumesList(PrettyList):
    def __init__(self, *args, **kwargs):
        super(MLPlatformVolumesList, self).__init__(args[0])


class VmInfo(BaseSerializerModel):
    id: str
    name: str
    status: str
    flavor_id: str
    created_dt: datetime.datetime
    public_ip: Optional[str] = None
    private_ip: Optional[str] = None
    domain_name: Optional[str] = None
    status_reason: Optional[str] = None
    instance_type: str
    mlflow_jh_instance_id: Optional[str] = None
    mlflow_deploy_instance_id: Optional[str] = None

    def __str__(self):
        attrs = vars(self)
        return '\n'.join("%s: %s" % item for item in attrs.items())


class VmDetailInfo(VmInfo):
    availability_zone: str
    volumes: MLPlatformVolumesList[MLPlatformVolumeIn]

    def __str__(self):
        attrs = vars(self)
        return ''.join("\n    %s: %s" % item for item in attrs.items())


class K8sClusterInfo(BaseSerializerModel):
    id: str
    name: str
    k8s_id: Optional[str] = None
    status: str
    status_reason: Optional[str] = None
    s3_bucket_name: Optional[str] = None
    docker_registry_id: str
    history_server_url: Optional[str] = None
    inactive_min: Optional[int] = None
    suspend_after_inactive_min: Optional[int] = None
    delete_after_inactive_min: Optional[int] = None
    docker_registry: Optional[VmDetailInfo] = None

    def __str__(self):
        attrs = vars(self)
        items = ''.join("\n%s: %s" % item for item in attrs.items())
        return f'\n---------------\nK8sClusterInfo:\n---------------{items}'


class K8sSparkJobLogs(BaseSerializerModel):
    logs: str

    def __str__(self):
        return self.logs


class K8sSparkEventInvolvedObject(BaseSerializerModel):
    api_version: Optional[str] = None
    field_path: Optional[str] = None
    kind: Optional[str] = None
    name: Optional[str] = None
    namespace: Optional[str] = None
    resource_version: Optional[str] = None
    uid: Optional[str] = None


class K8sSparkEventManagedFields(BaseSerializerModel):
    api_version: Optional[str] = None
    fields_type: Optional[str] = None
    fields_v1: Optional[Dict] = None
    manager: Optional[str] = None
    operation: Optional[str] = None
    subresource: Optional[str] = None
    time: Optional[datetime.datetime] = None


class K8sSparkEventMetodata(BaseSerializerModel):
    annotations: Optional[str] = None
    cluster_name: Optional[str] = None
    creation_timestamp: Optional[datetime.datetime] = None
    deletion_grace_period_seconds: Optional[str] = None
    deletion_timestamp: Optional[str] = None
    finalizers: Optional[str] = None
    generate_name: Optional[str] = None
    generation: Optional[str] = None
    labels: Optional[str] = None
    managed_fields: List[K8sSparkEventManagedFields]
    name: Optional[str] = None
    namespace: Optional[str] = None
    owner_references: Optional[str] = None
    resource_version: Optional[str] = None
    self_link: Optional[str] = None
    uid: Optional[str] = None


class K8sSparkEventSource(BaseSerializerModel):
    component: Optional[str] = None
    host: Optional[str] = None


class K8sSparkEventData(BaseSerializerModel):
    action: Optional[str] = None
    api_version: Optional[str] = None
    count: Optional[int] = None
    event_time: Optional[datetime.datetime] = None
    first_timestamp: Optional[datetime.datetime] = None
    involved_object: Optional[K8sSparkEventInvolvedObject] = None
    kind: Optional[str] = None
    last_timestamp: Optional[datetime.datetime] = None
    message: Optional[str] = None
    metadata: Optional[K8sSparkEventMetodata] = None
    reason: Optional[str] = None
    related: Optional[str] = None
    reporting_component: Optional[str] = None
    reporting_instance: Optional[str] = None
    series: Optional[str] = None
    source: K8sSparkEventSource
    type: Optional[str] = None


class K8sSparkEvent(BaseSerializerModel):
    event: Optional[K8sSparkEventData] = None

    @field_validator("event", mode="before")
    @classmethod
    def validate_event(cls, v: Any) -> Optional[K8sSparkEventData]:
        if isinstance(v, str):
            return K8sSparkEventData(**eval(v))
        if isinstance(v, K8sSparkEventData):
            return v
        return None

    def __str__(self):
        event = getattr(self, "event")
        if event is None:
            return ""
        involved_object = getattr(event, "involved_object")

        first_timestamp = event.first_timestamp.strftime('%Y/%m/%d %H:%M:%S') if getattr(event, "first_timestamp") else ""
        involved_object_name = event.involved_object.name if involved_object and getattr(involved_object, "name") else "unknown"
        event_type = event.type.upper() if getattr(event, "type") else ""
        involved_object_kind = event.involved_object.kind if involved_object and getattr(involved_object, "kind") else ""
        event_message = event.message if getattr(event, "message") else ""

        return f"{first_timestamp} Job={involved_object_name} {event_type} {involved_object_kind} {event_message}"


class K8sSparkEventsList(PrettyList):
    def __init__(self, *args, **kwargs):
        super(K8sSparkEventsList, self).__init__(args[0])


class K8sSparkApplication(BaseSerializerModel):
    apiVersion: str
    kind: str
    metadata: Optional[dict] = None
    spec: Optional[dict] = None
    status: Optional[dict] = None
    events: Optional[List[K8sSparkEvent]] = None

    def __str__(self):
        return safe_dump(self.not_null_values())


class K8sBaseConfigModel(BaseSerializerModel):
    name: str
    namespace: str
    labels: Optional[Dict[str, str]] = None
    annotations: Optional[Dict[str, str]] = None
    data: Optional[Dict[str, str]] = None

    def __str__(self):
        return safe_dump(self.not_null_values())


class K8sSecret(K8sBaseConfigModel):
    type: Optional[str] = None


class K8sConfigmap(K8sBaseConfigModel):
    binary_data: Optional[Dict[str, str]] = None


class K8sObjectNamesList(PrettyList):
    def __init__(self, *args, **kwargs):
        super(K8sObjectNamesList, self).__init__(args[0])


class K8sManifestIn(BaseSerializerModel):
    manifest: str


class K8sInactiveTimeoutIn(BaseSerializerModel):
    suspend_timeout_min: Optional[int] = None
    delete_timeout_min: Optional[int] = None

    @field_validator("suspend_timeout_min")
    @classmethod
    def validate_suspend_timeout(cls, value):
        min_value = 10
        max_value = 4320
        if value is not None and value != 0 and not min_value <= value <= max_value:
            raise ValueError(
                f"Suspend timeout={value} should be in range from {min_value} to {max_value}"
            )
        return value

    @field_validator("delete_timeout_min")
    @classmethod
    def validate_delete_timeout(cls, value):
        min_value = 30
        max_value = 10080
        if value is not None and value != 0 and not min_value <= value <= max_value:
            raise ValueError(
                f"Delete timeout={value} should be in range from {min_value} to {max_value}"
            )
        return value
