from pyspark import SparkFiles
from pyspark.sql import SparkSession

from pyspark.ml.regression import GBTRegressor
from pyspark.ml.evaluation import RegressionEvaluator
from pyspark.ml.feature import VectorAssembler
from pyspark.ml.tuning import ParamGridBuilder, CrossValidator
from pyspark.sql.functions import *
from pyspark.sql.types import *

spark = SparkSession.builder.appName("GradientBoostingExample").getOrCreate()


df = spark.read.csv("BostonHousing.csv", header=True, inferSchema=True)
df.show(5)

# Define the feature and label columns & Assemble the feature vector
assembler = VectorAssembler(inputCols=df.columns[:-1], outputCol='features')
data = assembler.transform(df)

# Split the data into training and test sets
train_data, test_data = data.randomSplit([0.7, 0.3], seed=42)

# Create a Gradient Boosting model
gbt = GBTRegressor(featuresCol='features', labelCol='medv', seed=42)

# Train the model
model = gbt.fit(train_data)

# Make predictions on the testing data
predictions = model.transform(test_data)

# Evaluate the model
rmse_eval = RegressionEvaluator(labelCol='medv', metricName='rmse')
mae_eval = RegressionEvaluator(labelCol='medv', metricName='mae')

rmse = rmse_eval.evaluate(predictions)
mae = mae_eval.evaluate(predictions)


print("RMSE: {:.2f}".format(rmse))
print("MAE: {:.2f}".format(mae))

# Define the hyperparameter grid
param_grid = ParamGridBuilder() \
    .addGrid(gbt.maxDepth, [2, 4, 6]) \
    .addGrid(gbt.maxIter, [10, 50, 100]) \
    .addGrid(gbt.stepSize, [0.1, 0.01]) \
    .build()

# Evaluate the model
evaluator = RegressionEvaluator(labelCol='medv', metricName='rmse')

# Define the cross-validator
crossval = CrossValidator(estimator=gbt,
                          estimatorParamMaps=param_grid,
                          evaluator=evaluator,
                          numFolds=5, seed=42)

# Train the model using cross-validation
cv_model = crossval.fit(train_data)

# Make predictions on the testing data
cv_predictions = cv_model.transform(test_data)

cv_rmse = evaluator.evaluate(cv_predictions)
print("CV RMSE: {:.2f}".format(cv_rmse))

# Get feature importance scores
importances = model.featureImportances

# Create a list of feature names
features = data.columns[:-1]

# Print the feature importance scores
for feature, importance in zip(features, importances):
    print(feature, "{:.4f}".format(importance))
