import logging
import time
from typing import Optional

from mlplatform_client import MLPlatform, MLPTokenType
import mlplatform_client.core.exceptions as client_exc
from mlplatform_client.serializers.spark_proxy import K8sSparkJobRestartPolicyType

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [ %(levelname)-7.7s ]  %(message)s",
    handlers=[logging.StreamHandler()],
)
log = logging.getLogger()


ADMIN_REFRESH_TOKEN = "<TOKEN>"
CLUSTER_ID = "<CLUSTER_ID>"


def wait_job(ml_client: MLPlatform, job_name: str, timeout: int = 300):
    log.info(f"({job_name:40s}) Waiting...")

    for i in range(timeout):
        try:
            job_info = ml_client.spark_job_info(CLUSTER_ID, job_name)
        except Exception:
            pass
        else:
            if job_info.pod_status == 'Succeeded':
                log.info(f"({job_name:40s}) Completed")
                return
        time.sleep(1)

    log.error(f"({job_name:40s}) ... Timeout")


def delete_job(ml_client: MLPlatform, job_name: str):
    log.info(f"({job_name:40s}) Deleting existing spark job")
    ml_client.spark_delete_job(CLUSTER_ID, job_name)

    try:
        _ = ml_client.spark_job_info(CLUSTER_ID, job_name)
    except client_exc.NotFoundException:
        log.info(f"({job_name:40s}) Spark job has been deleted")


def get_job_logs(ml_client: MLPlatform, job_name: str) -> Optional[str]:
    try:
        logs = ml_client.spark_job_logs(cluster_id=CLUSTER_ID, job_name=job_name)
        log.info(f"Job {job_name} logs:")
        log.info(logs)
    except Exception as e:
        log.error(f"({job_name:40s}) Getting logs error: {str(e)}")
        return None


def get_events(ml_client: MLPlatform) -> Optional[str]:
    try:
        events = ml_client.spark_events(cluster_id=CLUSTER_ID)
        log.info(f"[CLUSTER] ({CLUSTER_ID:40s}) events:")
        for event in events:
            log.info(event)
    except Exception as e:
        log.error(f"[CLUSTER] ({CLUSTER_ID:40s}) Getting events: {str(e)}")
        return None


def pipeline(ml_client: MLPlatform):
    manifest = ml_client.get_default_manifest(CLUSTER_ID)
    manifest.set_restart_policy(
        restart_type=K8sSparkJobRestartPolicyType.ON_FAILURE,
        on_failure_retries=3,
        on_failure_retry_interval=20,
        on_submission_failure_retries=5,
        on_submission_failure_retry_interval=20,
    )
    job_c_name = manifest.job_name

    log.info(f"({job_c_name:40s}) Submit spark job with code...")

    job_c_info = ml_client.spark_submit_job(CLUSTER_ID, manifest, "jobs/custom_job.py")
    log.info(f"({job_c_name:40s}) ... {job_c_info.status}")

    wait_job(ml_client, job_c_name)

    get_job_logs(ml_client=ml_client, job_name=job_c_name)
    get_events(ml_client=ml_client)
    delete_job(ml_client, job_c_name)


def main():
    admin_client = MLPlatform(ADMIN_REFRESH_TOKEN)
    user_client = MLPlatform()

    user_token_name = "test_user_2"
    user_register_token = admin_client.create_register_token(
        client_name=user_token_name,
        token_type=MLPTokenType.USER,
        access_ttl="1h",
        refresh_ttl="2h",
    )
    user_refresh_token = user_client.create_refresh_token(user_register_token)
    log.warning(f"!!! User must remember these tokens for future usage !!!")
    log.warning(f"User refresh token: {user_refresh_token}")

    try:
        _ = user_client.create_register_token(
            client_name="test_user_2",
            access_ttl="1h",
            refresh_ttl="2h",
        )
    except client_exc.ForbiddenException:
        log.warning(f"Users cannot create other users")

    log.info(f"Admin token info: {admin_client.who_am_i()}")
    log.info(f"User token info: {user_client.who_am_i()}")
    log.info(f"Project tokens info: {admin_client.get_project_tokens_info()}")

    pipeline(admin_client)
    pipeline(user_client)

    admin_client.delete_token(token_name=user_token_name)

    admin_client.stop()
    user_client.stop()


if __name__ == "__main__":
    main()
